from __future__ import annotations

import requests
from typing import Optional, Type
from datetime import datetime, timedelta

import wxcadm.person
from wxcadm import log
from .common import *
from .exceptions import *
from .org import Org
from .person import Me, Person


class Webex:
    """The base class for working with wxcadm"""
    def __init__(self,
                 access_token: str,
                 get_locations: bool = True,
                 get_xsi: bool = False,
                 get_hunt_groups: bool = False,
                 get_call_queues: bool = False,
                 fast_mode: bool = False,
                 client_id: Optional[str] = None,
                 client_secret: Optional[str] = None,
                 refresh_token: Optional[str] = None,
                 org_id: Optional[str] = None,
                 auto_refresh_token: bool = False
                 ) -> None:
        """Initialize a Webex instance to communicate with Webex and store data

        Args:
            access_token (str): The Webex API Access Token to authenticate the API calls
            get_locations (bool, optional): Whether to get all Locations and create instances for them. Defaults to
                True when there is only one Org. When more than one Org is present, setting this value to True has
                no effect and the Org-level method must be used.
            get_xsi (bool, optional): Whether to get the XSI endpoints for each Org. Defaults to False, since
                not every Org has XSI capability
            get_hunt_groups (bool, optional): Whether to get the Hunt Groups for each Org. Defaults to False. Setting
                this value to True only applies when one Org is present. If more than one Org is present, this arg
                is ignored and the Org-level method must be used.
            get_call_queues (bool, optional): Whether to get the Call Queues for each Org. Defaults to False. Setting
                this value to True only applies when one Org is present. If more than one Org is present, this arg
                is ignored and the Org-level method must be used.
            fast_mode (bool, optional): When possible, optimize the API calls to Webex to work more quickly,
                sometimes at the expense of not getting as much data. Use this option only if you have a script that
                runs very slowly, especially during the Webex initialization when collecting people. **Note that this
                option should not be used when it is necessary to know the phone numbers of each Person, because
                it skips the API call to the Call Control back-end on initialization.**
            client_id (str, optional): The Client ID or Application ID to associate with the token. This value is only
                useful if you are planning to call the :py:meth:`refresh_token()` method to refresh the token.
            client_secret (str, optional): The Client Secret for the Integration or Service Application. This value is
                only useful if you are planning to call the :py:meth:`refresh_token()` method to refresh the token.
            refresh_token (str, optional): The Refresh Token associated with the Access Token. This argument is needed
                if you are planning to call the :py:meth:`refresh_token()` method to refresh the Access Token.
            org_id (str, optional): The Org ID to use as the default Org for the session. Other Orgs are still available
                in the :attr:`Webex.orgs` list.
            auto_refresh_token (bool, optional): When True, any API call will first check the access token expiration
                and automatically refresh the token if the expiration is within the next 30 minutes. Note that this
                requires the ``client_id``, ``client_secret`` and ``refresh_token`` to be provided when the
                :class:`Webex` instance is created, as those values are needed by the refresh process. **This feature
                is still in development and should not be used until this warning is removed**

        Returns:
            Webex: The Webex instance

        """
        log.info("Webex instance initialized")
        # The access token is the only thing that we need to get started
        self._access_token: str = access_token
        # The Authorization header is going to be used by every API call in the package.
        # Might want to make it something global, so we don't have to inherit it across all the children
        self._headers: dict = {"Authorization": "Bearer " + access_token}
        log.debug(f"Setting Org._headers to {self._headers}")
        log.debug(f"Setting Global _webex_headers")
        global _webex_headers
        _webex_headers['Authorization'] = "Bearer " + access_token

        # Fast Mode flag when needed
        self._fast_mode = fast_mode

        # Instance attrs
        self.client_id = client_id
        """ The Client ID or Application ID """
        self.client_secret = client_secret
        """ The Client Secret value for the Integration or Service Application """
        self.refresh_token = refresh_token
        """ The Refresh Token associated with the Access Token """
        self.access_token_expires = None
        """ The datetime when the access token expires """
        self.refresh_token_expires = None
        """ The datetime when the refresh token expires """
        self.auto_refresh_token: bool = auto_refresh_token

        self.orgs: list = []
        '''A list of the Org instances that this Webex instance can manage'''
        self.org: Optional[Org] = None
        """
        If there is only one Org in :py:attr:`Webex.orgs` or if the ``org_id`` param was passed, this attribute will be
        the first or selected Org.
        """
        self._me: Optional[Type[Me]] = None
        # Get the orgs that this token can manage
        log.debug(f"Making API call to v1/organizations")
        r = requests.get(_url_base + "v1/organizations", headers=self._headers)
        # Handle invalid access token
        if r.status_code != 200:
            log.critical("The Access Token was not accepted by Webex")
            log.debug(f"Response: {r.text}")
            raise TokenError("The Access Token was not accepted by Webex")
        response = r.json()
        # Handle when no Orgs are returned. This is pretty rare
        if len(response['items']) == 0:
            log.warning("No Orgs were returned by the Webex API")
            raise OrgError

        orgs = response['items']
        for org in orgs:
            log.debug(f"Processing org: {org['displayName']}")
            this_org = Org(name=org['displayName'], id=org['id'], parent=self, xsi=False)
            self.orgs.append(this_org)
        if org_id is not None:
            log.info(f"Setting Org ID {org_id} as default Org")
            self.org = self.get_org_by_id(org_id)
            if self.org is None:
                log.warning("Org not found")
                raise OrgError
        else:
            self.org = self.orgs[0]


    @property
    def headers(self):
        """The "universal" HTTP headers with the Authorization header present"""
        return self._headers

    def get_new_token(self, client_id: Optional[str] = None,
                      client_secret: Optional[str] = None,
                      refresh_token: Optional[str] = None):
        """ Refresh the Access Token

        To perform a refresh, you must know the client_id and client_secret, and refresh_token value. If you have not
        already set the attributes on the :py:class:`Webex` instance, you can provide them as arguments to this method.
        Passing those arguments will set the instance attributes, so they don't have to be passed on every call to this
        method, meaning any existing value will be overwritten.

        When the Access Token is refreshed, the :py:class:`Webex` instance will be updated to use the new token.

        Args:
            client_id (str): The Client ID or Application ID of your Integration or Service Application
            client_secret (str): The Client Secret of your Integration or Service Application
            refresh_token (str): The Refresh Token associated with the given Access Token

        Returns:
            dict: The new token information

        """
        if client_id is None:
            if self.client_id is None:
                raise ValueError
        else:
            self.client_id = client_id
        if client_secret is None:
            if self.client_secret is None:
                raise ValueError
        else:
            self.client_secret = client_secret
        if refresh_token is None:
            if self.refresh_token is None:
                raise ValueError
        else:
            self.refresh_token = refresh_token

        payload = {
            'grant_type': 'refresh_token',
            'client_id': self.client_id,
            'client_secret': self.client_secret,
            'refresh_token': self.refresh_token
        }

        response = webex_api_call('post', '/v1/access_token', payload=payload)
        if 'access_token' in response:
            log.info(f"Changing Access Token to {response['access_token']}")
            self._access_token = response['access_token']
            self._headers: dict = {"Authorization": "Bearer " + self._access_token}

            log.debug(f"Setting Org._headers to {self._headers}")
            log.debug(f"Setting Global _webex_headers")
            global _webex_headers
            _webex_headers['Authorization'] = "Bearer " + self._access_token

            log.debug("Setting token expiry for Org instance")
            self.access_token_expires = datetime.now() + timedelta(seconds=response['expires_in'])
            self.refresh_token_expires = datetime.now() + timedelta(seconds=response['refresh_token_expires_in'])
            return response
        else:
            return False

    def get_org_by_name(self, name: str):
        """Get the Org instance that matches all or part of the name argument.

        Args:
            name (str): Text to match against the Org name

        Returns:
            Org: The Org instance of the matching Org

        Raises:
            wxcadm.exceptions.KeyError: Raised when no match is made

        """
        for org in self.orgs:
            if name.lower() in org.name.lower():
                return org
        raise KeyError("Org not found")

    def get_org_by_id(self, id: str):
        """Get the Org instance by Org ID.

        Args:
            id (str): The ID of the Org to find

        Returns:
            Org: The Org instance of the matching Org

        Raises:
            wxcadm.exceptions.KeyError: Raised when no match is made

        """
        for org in self.orgs:
            if org.id == id or org.spark_id.split("/")[-1] == id:
                return org
        raise KeyError("Org not found")

    def get_person_by_email(self, email: str) -> Optional[wxcadm.person.Person]:
        """ Get the person instance  of a user with the given email address

        Unlike the :class:`Org` method of the same name, this method searches across all the Orgs that the token
        has access to, so it can find a user in any :class:`Org`

        Args:
            email (str): The email address to search for

        Returns:
            :class:`Person`: The Person instance. None is returned if no match is found

        """
        log.info(f"Getting Person record for email: {email}")
        response = webex_api_call("get", "v1/people", params={"email": email, "callingData": True})
        if len(response) > 1:
            log.warn(f"Webex returned more than one record for email: {email}")
            raise APIError("Webex returned more than one Person with the specified email")
        elif len(response) == 1:
            log.debug(f"User data: {response[0]}")
            org = self.get_org_by_id(response[0]['orgId'])
            log.debug(f"User in Org: {org.name}")
            person = Person(response[0]['id'], parent=org, config=response[0])
            return person
        else:
            return None

    def get_person_by_id(self, id: str):
        """ Get the Person instance for a user with the given ID

        Unlike the :class:`Org` method of the same name, this method searches across all Orgs that the token has
        access to, so it can find a user in any :class:`Org`

        Args:
            id (str): The ID to search for

        Returns:
            Person: The Person instance. None is returned if no match is found

        """
        log.info(f"Getting Person record for ID: {id}")
        response = webex_api_call("get", "v1/people", params={"id": id, "callingData": True})
        if len(response) > 1:
            log.warn(f"Webex returned more than one record for id: {id}")
            raise APIError("Webex returned more than one Person with the specified ID")
        elif len(response) == 1:
            log.debug(f"User data: {response[0]}")
            org = self.get_org_by_id(response[0]['orgId'])
            log.debug(f"User in Org: {org.name}")
            person = Person(response[0]['id'], parent=org, config=response[0])
            return person
        else:
            return None

    @property
    def me(self):
        """ An instance of the :py:class:`Me` class representing the token owner """
        if self._me is None:
            my_info = webex_api_call("get", "v1/people/me", headers=self.headers)
            me = Me(my_info['id'], parent=self.get_org_by_id(my_info['orgId']), config=my_info)
            self._me = me
        return self._me



from __future__ import annotations
from collections import UserList
from typing import TYPE_CHECKING
import json
from dataclasses import dataclass
from dataclasses_json import dataclass_json, LetterCase
import logging
import wxcadm
from .common import *
from typing import Optional, Union
from .exceptions import *
from wxcadm import log
from .virtual_line import VirtualLine
if TYPE_CHECKING:
    from .person import Person
    from .workspace import Workspace


class DeviceLayout:
    def __init__(self, device: Device, config: Optional[dict] = None):
        self.device = device
        self.layout_mode: str = ''
        self.user_reorder_enabled: bool = False
        self.line_keys: list = []
        self.kem_type: Optional[str] = None
        self.kem_keys: Optional[list] = None

        if config is None:
            self._get_config()
        else:
            self._parse_config(config)

    def _parse_config(self, config: dict):
        self.layout_mode = config.get('layoutMode', '')
        self.user_reorder_enabled = config.get('userReorderEnabled', False)
        self.line_keys = config.get('lineKeys', [])
        self.kem_type = config.get('kemModuleType', None)
        self.kem_keys = config.get('kemKeys', None)

    def _get_config(self):
        response = webex_api_call('get', f'v1/telephony/config/devices/{self.device.id}/layout',
                                  params={'orgId': self.device.parent.org_id})
        self._parse_config(response)


class Device:
    def __init__(self, parent: wxcadm.Org | Person | Workspace,
                 config: Optional[dict] = None,
                 id: Optional[str] = None):
        self.parent = parent
        """ The :py:class:`Org`, :py:class:`Person` or :py:class:`Workspace` that created this instance """

        # The logic below addresses the case where the Device owner is a Person, because of how we have to get the
        # list of devices
        if "connectionStatus" not in config.keys():
            try:
                config = webex_api_call('get', f'/v1/devices/{id}', params={'orgId': self.parent.org_id})
            except wxcadm.APIError:
                pass

        self.id: str = config['id']
        """ The Device ID """
        self.tags: list = config.get('tags', config.get('description', []))
        """ The device tags """
        self.model: str = config.get('model', config.get('product', ''))
        """ The model name of the device """
        self.mac: str = config.get('mac')
        """ The device MAC address """
        self.is_owner: bool = config.get('primaryOwner', False)
        """ Whether the Person or Workspace is the primary owner of the device """
        self.activation_state: Optional[str] = config.get('activationState', None)
        """ The device activation state """
        self.type: str = config['type']
        """ The type of device association """
        self.ip_address: Optional[str] = config.get('ipAddress', config.get('ip', ''))
        """ The IP Address of the device """
        self._settings: Optional[dict] = None
        self.display_name = config.get('displayName', None)
        """ The name of device as displayed in Webex """
        self.capabilities: list = config.get('capabilities', [])
        """ The capabilities of the device """
        self.user_permissions: list = config.get('permissions', [])
        """ The permissions the user has for this device. For example, "xapi" means this user is allowed to use xapi """
        self.connection_status: Optional[str] = config.get('connectionStatus', None)
        """ Whether the device is connected or not """
        self.serial: Optional[str] = config.get('serial', None)
        """ The serial number of the device, if available """
        self.software: Optional[str] = config.get('software', None)
        """ The operating system name data and version tag, if available """
        self.upgrade_channel: Optional[str] = config.get('upgradeChannel', None)
        """ The upgrade channel the device is assigned to """
        self.created: Optional[str] = config.get('created', None)
        """ The date and time that the device was created on Webex """
        self.first_seen: Optional[str] = config.get('firstSeen', None)
        """ The date and time that the device was first seen online """
        self.last_seen: Optional[str] = config.get('lastSeen', None)
        """ The date and time that the device was last seen online """
        self.owner = None
        """ The :py:class:`Person` or :py:class:`Workspace` that owns the device primarily """
        self.workspace_location_id: Optional[str] = config.get('workspaceLocationId', None)
        """ The WorkspaceLocation ID of the device, which indicates the WorkspaceLocation of the primary
            Person or Workspace
            
            .. deprecated:: 3.4.0
            The Workspace Location concept is being removed from the Webex APIs. All Workspace Locations are now simply
            :class:`Location`s now. This class will still show up in come cases until it is completely removed. Use
            :attr:`Device.location_id` instead. 
            
        """
        self.location_id: Optional[str] = config.get('locationId', None)
        """ The Location ID of the device, which indicates the Location of the primary Person or Workspace """
        self._device_members = None
        self._layout = None

        if 'personId' in config.keys() and isinstance(self.parent, wxcadm.Org):
            self.owner = self.parent.get_person_by_id(config['personId'])
            if self.owner is None:
                self.owner = config['personId']
        elif 'workspaceId' in config.keys() and isinstance(self.parent, wxcadm.Org):
            self.owner = self.parent.workspaces.get_by_id(config['workspaceId'])

    @property
    def layout(self) -> DeviceLayout:
        """ The :class:`DeviceLayout` for the Device """
        if self._layout is None:
            self._layout = DeviceLayout(self)
        return self._layout

    def set_layout(self, layout: DeviceLayout):
        """ Use the provided :class:`DeviceLayout` as the new layout for the Device.

        .. note::

            Due to the complexity of the DeviceLayout, the lists and dicts for each line section are represented as
            they are defined by the Webex Developer documents. See the documented format at:
            https://developer.webex.com/docs/api/v1/device-call-settings/modify-device-layout-by-device-id

        Args:
            layout (DeviceLayout: The :class:`DeviceLayout` to apply

        Returns:
            bool: True on success

        """
        payload = {
            'layoutMode': layout.layout_mode,
            'userReorderEnabled': layout.user_reorder_enabled,
            'lineKeys': layout.line_keys
        }
        if layout.kem_type is not None:
            payload['kemModuleType'] = layout.kem_type
        if layout.kem_keys is not None:
            payload['kemKeys'] = layout.kem_keys

        webex_api_call('put', f"v1/telephony/config/devices/{self.id}/layout", payload=payload,
                       params={'orgId': self.parent.org_id})
        return True

    def change_tags(self, operation: str, tag: Optional[Union[str, list]] = None):
        """ Add a tag to the list of tags for this device

        It is also possible to change the tags by modifying the ``settings`` directly, but this method ensures that
        no other settings are touched.

        Args:
            operation (str): Valid values are:
                "add" - Add a tag or list of tags to the existing tags
                "remove" - Remove all tags for the device
                "replace" - Replace all tags with the provided tag or list of tags
            tag (str, list, optional): The tag value or list of values. Valid for "add" and "replace" operations.

        Returns:

        """
        # Value checking
        if (operation.lower() in ['add', 'replace']) and tag is None:
            raise ValueError("add or replace operations require a 'tag' param")
        if isinstance(tag, str):
            tag = [tag]

        payload = {
            "op": operation.lower(),
            "path": "tags",
            "value": tag
        }
        webex_api_call("patch", f"/v1/devices/{self.id}", payload=payload, params={'orgId': self.parent.org_id})

        return True

    @property
    def config(self) -> dict:
        response = webex_api_call('get', f'/v1/telephony/config/devices/{self.id}')
        return response

    @property
    def settings(self):
        """ The device settings.

        The available settings vary by device type, so the raw dictionary from Webex is returned.

        """
        if self._settings is None:
            response = webex_api_call('get', f'/v1/telephony/config/devices/{self.id}/settings',
                                      params={'model': self.model, 'orgId': self.parent.org_id})
            self._settings = response
        return self._settings

    @settings.setter
    def settings(self, config: dict):
        try:
            webex_api_call('put', f'v1/telephony/config/devices/{self.id}/settings',
                                  payload=config, params={'orgId': self.parent.org_id})
        except APIError:
            logging.warning("The API call to set the device settings failed")

        self._settings = config

    def apply_changes(self) -> bool:
        """ Issues request to the device to download and apply changes to the configuration.

        Returns:
            bool: True on success, False otherwise

        """
        try:
            webex_api_call('post', f'v1/telephony/config/devices/{self.id}/actions/applyChanges/invoke',
                           params={'orgId': self.parent.org_id})
        except APIError:
            return False

        return True

    def delete(self):
        """ Delete the device from Webex Calling and from Control Hub completely

        Returns:
            bool: True on success. An exception will be thrown otherwise

        """
        webex_api_call("delete", f"v1/devices/{self.id}", params={'orgId': self.parent.org_id})
        return True

    @property
    def members(self):
        """ :class:`DeviceMembersList` of the configured lines (i.e. members) on the device """
        if self._device_members is None:
            self._device_members = DeviceMemberList(self)
        return self._device_members

    def get_workspace_location_uuid(self):
        """ Returns just the UUID portion of the `workspace_location_id` property.

        .. versionadded:: 4.2.2
            Added because the Base64 ID does not match the ID values for :class:`WorkspaceLocation`. The assumption is
            that this will be resolved in the future, but needs to be supported until then

        Returns:
            str: The UUID component of the :attr:`workspace_location_id`

        """
        return decode_spark_id(self.workspace_location_id).split("/")[-1]


class DeviceMemberList(UserList):
    def __init__(self, device: Device):
        log.info(f"Collecting Device Members for {device.display_name}")
        super().__init__()
        log.info("Initializing DeviceMemberList")
        self.device = device
        self.max_line_count: Optional[int] = None
        self.data = self._get_data()

    def _get_data(self):
        try:
            response = webex_api_call('get', f"v1/telephony/config/devices/{self.device.id}/members",
                                      params={'orgId': self.device.parent.org_id})
        except:
            return []
        log.debug(response)
        members = []
        for entry in response['members']:
            members.append(DeviceMember(self.device, entry))
        self.max_line_count = response['maxLineCount']
        return members

    def refresh(self):
        """ Refresh the list of configured lines from Webex """
        self.data = self._get_data()

    def available_members(self) -> list:
        """ Get a list of available members, which are Workspaces and People that can be assigned to this device.

        This returns a dict, which is what is returned by Webex. I am not sure what the purpose of this API call is,
        but I see it used in Control Hub, so I am including it in case it provides some value.

        Returns:
            dict: A list of available members to add to the device

        """
        response = webex_api_call('get', f"v1/telephony/config/devices/{self.device.id}/availableMembers",
                                  params={'orgId': self.device.parent.org_id})
        return response['members']

    def add(self, members: Union[list, Workspace, Person, VirtualLine],
            line_type: str = 'shared',
            line_label: Optional[str] = None,
            hotline_enabled: bool = False,
            hotline_destination: Optional[str] = None,
            allow_call_decline: bool = False) -> bool:
        """ Add a new configured line (member) to the Device

        .. note::

            When using the ``line_type``, ``hotline_enabled`` or ``allow_call_decline`` arguments, those will be
            applied to all members in the list, if provided. If you need different settings for each member, call the
            add() method multiple times with the relevant settings.

        Args:
            members (list, Workspace, Person, VirtualLine): The Workspace, Person, VirtualLine or a list of those to
                add as configured lines.

            line_type (str, optional): Allowed values are ``'primary'`` for the primary device for a Person or Workspace
                or ``'shared'`` for Shared Line Appearances on non-primary devices. Defaults to ``'shared'``

            line_label (str, optional): A text label for the line on the device. Note that this is only supported for
                Cisco MPP devices. Attempting to set a line label on a Customer or Partner Managed Device will fail.

            hotline_enabled (bool, optional): Whether Hotline should be enabled for the line. Default False.
            hotline_destination (str, optional): The Hotline destination number. Required when ``hotline_enabled``
                is True

            allow_call_decline (bool, optional): Whether declining a call will decline across all devices or just
                silence this device. Defaults to False, which silences this device only.

        Returns:
            bool: True on success, False otherwise.

        """
        log.info(f"Adding to Device Members to {self.device.display_name}")
        ports_available = self.ports_available()
        members_list_json: list = self._json_list()
        # If a Workspace or Person was provided, put it into a list anyway
        if isinstance(members, (wxcadm.person.Person, wxcadm.workspace.Workspace, wxcadm.virtual_line.VirtualLine)):
            members = [members]
        # The following section was removed because port assignment doesn't seem to do anything, at least with MPP
        # if port is not None:
        #     # Check to make sure the port is available, or the whole port range
        #     for i in range(port, port + len(members), 1):
        #         if i not in ports_available:
        #             log.warning(f"Requested port {port} is already in use.")
        #             raise ValueError(f"Port {port} is already in use")

        port = ports_available[0]
        log.debug(f"Using port {port} as the starting port number")

        # Set some defaults to the right values for the API call
        line_type = 'PRIMARY' if line_type.lower() == 'primary' else 'SHARED_CALL_APPEARANCE'

        for new_member in members:
            member_config = {
                'port': port,
                'id': new_member.id,
                'primaryOwner': False,
                'lineType': line_type,
                'lineWeight': 1,
                'hotlineEnabled': hotline_enabled,
                'hotlineDestination': hotline_destination,
                'allowCallDeclineEnabled': allow_call_decline,
            }
            # lineLabel cannot be passed to non-MPP devices, so don't add it unless we need to
            if line_label is not None:
                member_config['lineLabel'] = line_label
            members_list_json.append(member_config)

        wxcadm.webex_api_call('put',
                              f"v1/telephony/config/devices/{self.device.id}/members",
                              payload={'members': members_list_json},
                              params={'orgId': self.device.parent.org_id})
        return True

    def _json_list(self) -> list:
        json_list = []
        entry: DeviceMember
        for entry in self.data:
            new_entry = {
                'port': entry.port,
                'id': entry.id,
                'primaryOwner': entry.primary_owner,
                'lineType': entry.line_type,
                'lineWeight': entry.line_weight,
                'hotlineEnabled': entry.hotline_enabled,
                'hotlineDestination': entry.hotline_destination,
                'allowCallDeclineEnabled': entry.call_decline_all,
            }
            # lineLabel is weird, so we have to only add it back when it is not None
            if entry.line_label is not None:
                new_entry['lineLabel'] = entry.line_label
            json_list.append(new_entry)
        return json_list

    def ports_available(self) -> list:
        """ Returns a list of available port numbers. """
        ports_available = []
        for port, member in self.port_map().items():
            if member is None:
                ports_available.append(port)
        return ports_available

    def port_map(self) -> dict:
        """ Returns a dict mapping of all ports and their assigned :class:`DeviceMember` """
        port_map = {}
        for i in range(1, self.max_line_count + 1, 1):
            port_map[i] = None
        member: DeviceMember
        for member in self.data:
            log.debug("Starting loop")
            for p in range(member.port, member.port + member.line_weight, 1):
                log.debug(f"In Loop: {p}")
                port_map[p] = member
        return port_map

    def get(self, person: Optional[Person] = None, workspace: Optional[Workspace] = None):
        if person is not None and workspace is not None:
            raise ValueError('Only person or workspace is accepted, not both')
        if person is not None:
            member_instance = person
        if workspace is not None:
            member_instance = workspace
        for member in self.data:
            if member.id is not None and (member.id == member_instance.id):
                return member
        return None


class DeviceMember:
    def __init__(self, device: Device, member_info: dict):
        self.device: Device = device
        self.member_type: str = member_info['memberType']
        self.id: str = member_info['id']
        self.port: int = member_info['port']
        self.primary_owner: bool = member_info['primaryOwner']
        self.line_type: str = member_info['lineType']
        self.line_weight: int = member_info['lineWeight']
        self.hotline_enabled: bool = member_info['hotlineEnabled']
        self.hotline_destination: Optional[str] = member_info.get('hotlineDestination', None)
        self.call_decline_all: bool = member_info['allowCallDeclineEnabled']
        self.line_label: Optional[str] = member_info.get('lineLabel', None)
        self.first_name: Optional[str] = member_info.get('firstName', None)
        self.last_name: Optional[str] = member_info.get('lastName', None)
        self.phone_number: Optional[str] = member_info.get('phoneNumber', None)
        self.extension: Optional[str] = member_info.get('extension', None)
        self.host_ip: Optional[str] = member_info.get('hostIP', None)
        self.remote_ip: Optional[str] = member_info.get('remoteIP', None)
        self.line_port: Optional[str] = member_info.get('linePort', None)

    def set_line_label(self, label: str) -> DeviceMember:
        """ Set/update the Line Label on the device

        Args:
            label (str): The new line label

        Returns:
            DeviceMember: The updated DeviceMember instance

        """
        # Rather than rebuild the member list from what we already have, it's quicker to just pull a fresh copy
        old_member_list = webex_api_call('get', f'v1/telephony/config/devices/{self.device.id}/members',
                                         params={'orgId': self.device.parent.org_id})
        new_member_list = []
        for member in old_member_list['members']:
            if member['id'] == self.id:
                member['lineLabel'] = label
            # The PUT doesn't take all the fields that the GET has so we have to clean up
            member.pop('remoteIP', None)
            member.pop('linePort', None)
            member.pop('firstName', None)
            member.pop('lastName', None)
            member.pop('phoneNumber', None)
            member.pop('extension', None)
            member.pop('hostIP', None)
            member.pop('location', None)

            new_member_list.append(member)

        # Once we have rebuilt the list, just PUT it back
        webex_api_call('put', f'v1/telephony/config/devices/{self.device.id}/members',
                       payload={'members': new_member_list}, params={'orgId': self.device.parent.org_id})
        self.line_label = label
        return self

    def set_hotline(self, enabled: bool, destination: Optional[str] = None) -> DeviceMember:
        """ Enable or disable Hotline for the Device Member

        Args:
            enabled (bool): True for Enabled, False for Disabled
            destination (str, optional): The Hotline destination

        Returns:
            DeviceMember: The updated DeviceMember instance

        """
        # Rather than rebuild the member list from what we already have, it's quicker to just pull a fresh copy
        old_member_list = webex_api_call('get', f'v1/telephony/config/devices/{self.device.id}/members',
                                         params={'orgId': self.device.parent.org_id})
        new_member_list = []
        for member in old_member_list['members']:
            if member['id'] == self.id:
                member['hotlineEnabled'] = enabled
                if destination is not None:
                    member['hotlineDestination'] = destination

            # The PUT doesn't take all the fields that the GET has so we have to clean up
            member.pop('remoteIP', None)
            member.pop('linePort', None)
            member.pop('firstName', None)
            member.pop('lastName', None)
            member.pop('phoneNumber', None)
            member.pop('extension', None)
            member.pop('hostIP', None)
            member.pop('location', None)

            new_member_list.append(member)

        # Once we have rebuilt the list, just PUT it back
        webex_api_call('put', f'v1/telephony/config/devices/{self.device.id}/members',
                       payload={'members': new_member_list}, params={'orgId': self.device.parent.org_id})
        self.hotline_enabled = enabled
        if destination is not None:
            self.hotline_destination = destination
        return self

    def set_call_decline_all(self, enabled: bool):
        """ Set or change the Call Decline behavior of the Device Member.

        When set to True, declining the call on the device will decline the call on all devices. When set to False,
        declining the call on the device will only silence the device and allow other devices to keep ringing.

        Args:
            enabled (bool): True to decline on all devices, False to decline on this device

        Returns:
            DeviceMember: The updated DeviceMember instance

        """
        # Rather than rebuild the member list from what we already have, it's quicker to just pull a fresh copy
        old_member_list = webex_api_call('get', f'v1/telephony/config/devices/{self.device.id}/members',
                                         params={'orgId': self.device.parent.org_id})
        new_member_list = []
        for member in old_member_list['members']:
            if member['id'] == self.id:
                member['allowCallDeclineEnabled'] = enabled

            # The PUT doesn't take all the fields that the GET has so we have to clean up
            member.pop('remoteIP', None)
            member.pop('linePort', None)
            member.pop('firstName', None)
            member.pop('lastName', None)
            member.pop('phoneNumber', None)
            member.pop('extension', None)
            member.pop('hostIP', None)
            member.pop('location', None)

            new_member_list.append(member)

        # Once we have rebuilt the list, just PUT it back
        webex_api_call('put', f'v1/telephony/config/devices/{self.device.id}/members',
                       payload={'members': new_member_list}, params={'orgId': self.device.parent.org_id})
        self.call_decline_all = enabled
        return self


class DeviceList(UserList):
    _endpoint = "v1/devices"
    _endpoint_items_key = None
    _item_endpoint = "v1/devices/{item_id}"
    _item_class = Device

    def __init__(self, parent: Union[wxcadm.Org, wxcadm.Location, wxcadm.Workspace, wxcadm.Person]):
        super().__init__()
        log.debug("Initializing DeviceList")
        self.parent: wxcadm.Org | wxcadm.person.Person | wxcadm.workspace.Workspace | wxcadm.Location = parent
        self.data: list = self._get_data()
        self._supported_devices: Optional[SupportedDeviceList] = None

    def _get_data(self) -> list:
        log.debug("_get_data() started")
        params = {}

        if isinstance(self.parent, wxcadm.Org):
            log.debug(f"Using Org ID {self.parent.id} as data filter")
            params['orgId'] = self.parent.id
        elif isinstance(self.parent, wxcadm.Location):
            log.debug(f"Using Location ID {self.parent.id} as data filter")
            params['locationId'] = self.parent.id
        elif isinstance(self.parent, wxcadm.person.Person):
            log.debug(f"Using API endpoint v1/telephony/config/people/{self.parent.id}/devices")
            self._endpoint = f"v1/telephony/config/people/{self.parent.id}/devices"
            self._endpoint_items_key = 'devices'
        elif isinstance(self.parent, wxcadm.workspace.Workspace):
            log.debug(f"Using Workspace ID {self.parent.id} as data filter")
            params['workspaceId'] = self.parent.id
        else:
            log.warn("Parent class is not Org or Location, so all items will be returned")
        response = webex_api_call('get', self._endpoint, params=params)
        items = []
        if self._endpoint_items_key is not None:
            log.info(f"Found {len(response[self._endpoint_items_key])} items")
            for entry in response[self._endpoint_items_key]:
                items.append(self._item_class(parent=self.parent, config=entry, id=entry['id']))
        else:
            log.info(f"Found {len(response)} items")
            for entry in response:
                items.append(self._item_class(parent=self.parent, config=entry, id=entry['id']))
        return items

    def refresh(self):
        """ Refresh the list of instances from Webex

        Returns:
            bool: True on success, False otherwise.

        """
        self.data = self._get_data()
        return True

    def get(self,
            id: Optional[str] = None,
            name: Optional[str] = None,
            mac_address: Optional[str] = None,
            spark_id: Optional[str] = None,
            connection_status: Optional[str] = None):
        """ Get the instance associated with a given ID, Name, or Spark ID

        Only one parameter should be supplied in normal cases. If multiple arguments are provided, the Locations will be
        searched in order by ID, Name, and finally Spark ID. If no arguments are provided, the method will raise an
        Exception.

        Args:
            id (str, optional): The Device ID to find
            name (str, optional): The Device Name to find. Case-insensitive.
            mac_address (str, optional): The Device MAC address
            spark_id (str, optional): The Spark ID to find
            connection_status(str, optional): The connection status of the device (e.g. "disconnected", "connected")

        Returns:
            Device: The Device instance, or list of instances correlating to the given search argument.

        Raises:
            ValueError: Raised when the method is called with no arguments

        """
        if id is None and name is None and mac_address is None and spark_id is None and connection_status is None:
            raise ValueError("A search argument must be provided")
        if id is not None:
            for item in self.data:
                if item.id == id:
                    return item
        if name is not None:
            for item in self.data:
                if item.display_name.lower() == name.lower():
                    return item
        if mac_address is not None:
            for item in self.data:
                if item.mac == mac_address.upper().replace(':', '').replace('-', ''):
                    return item
        if spark_id is not None:
            for item in self.data:
                if item.spark_id == spark_id:
                    return item
        if connection_status is not None:
            item_list: list = []
            for item in self.data:
                if item.connection_status.lower() == connection_status.lower():
                    item_list.append(item)
            return item_list
        return None

    def create(self, model: SupportedDevice,
               mac: Optional[str] = None,
               password: Optional[str] = None,
               person: Optional[wxcadm.person.Person] = None,
               workspace: Optional[wxcadm.workspace.Workspace] = None):
        """ Add a new device to the Workspace, Person or Org

        In order to use this method, you must know the model of the device that you are adding, which is a
        :class:`SupportedDevice` from the :attr:`Org.supported_devices` :class:`SupportedDeviceList`.

        If the MAC address is passed, a device will be created with the provided MAC address. If no MAC address is
        passed, an Activation Code will be generated and returned as part of the response. Your integration/token must
        have the ``identity:placeonetimepassword_create`` scope to create Activation Codes for devices. Some supported
        devices do not allow onboarding by Activation Code, which will raise a ValueError. You can check the
        :attr:`SupportedDevice.onboarding_method` attribute to determine what onboarding methods are allowed.

        Args:
            model (SupportedDevice): The :class:`SupportedDevice` model that is being added

            mac (str, optional): The MAC address of the device being added

            password (str, optional): Only valid when creating devices that are not managed by Cisco. If a
                password is not provided, the Webex API will generate a unique, compliant SIP password and return it
                in the response.

            person (Person, optional): If the :class:`DeviceList` was accessed via :attr:`Org.devices`, a Person or
                Workspace must be provided when creating a Device

            workspace (Workspace, optional): If the :class:`DeviceList` was accessed via :attr:`Org.devices`, a
                Workspace or Person must be provided when creating a Device

        Returns:
            dict: The dict values vary based on the type of device being activated. If the device fails for any reason,
                False will be returned. At the moment, Webex doesn't provide very useful failure reasons, but those may
                be added to the return value when they are available.

        Raises:
            ValueError: Raised when the DeviceList cannot determine which Person or Workspace to add the device to.
                This is normally when the DeviceList was created at the Org level with :attr:`Org.devices`. Ensure you
                are passing a ``workspace`` or ``person`` argument to the method. It may also be raised if the device
                type does not support the requested onboarding method.

        """
        log.info(f"Adding a device to {self.parent.name}")
        payload = {}
        if isinstance(model, SupportedDevice):
            log.debug(f"Checking SupportedDevice requirements for {model.model}")
            log.debug(f"Setting model name to {model.model}")
            payload['model'] = model.model
            if model.managed_by.upper() == 'CISCO':
                log.debug("Cisco-managed device. No password needed.")
                password_needed = False
                data_needed = False
            else:
                log.debug(f"{model.managed_by.title()}-managed device. Password needed.")
                password_needed = True
                data_needed = True
            if 'ACTIVATION_CODE' not in model.onboarding_method and mac is None:
                raise ValueError("Activation Code not supported")
            if isinstance(self.parent, wxcadm.Workspace) or workspace is not None:
                if 'PLACE' not in model.supported_for:
                    raise ValueError("Device does not support Workspaces")
            if isinstance(self.parent, wxcadm.Person) or person is not None:
                if 'PEOPLE' not in model.supported_for:
                    raise ValueError("Device does not support People")
        else:
            raise ValueError("model argument must be a SupportedDevice object")

        if isinstance(self.parent, wxcadm.workspace.Workspace):
            payload['workspaceId'] = self.parent.id
            params = {'orgId': self.parent.org_id}
            device_info_url = f"v1/telephony/config/workspaces/{self.parent.id}/devices"
        elif isinstance(self.parent, (wxcadm.person.Person, wxcadm.person.Me)):
            payload['personId'] = self.parent.id
            params = {'orgId': self.parent.org_id}
            device_info_url = f"v1/telephony/config/people/{self.parent.id}/devices"
        elif isinstance(self.parent, wxcadm.org.Org):
            if person is not None:
                payload['personId'] = person.id
                params = {'orgId': self.parent.id}
                device_info_url = f"v1/telephony/config/people/{person.id}/devices"
            elif workspace is not None:
                payload['workspaceId'] = workspace.id
                params = {'orgId': self.parent.id}
                device_info_url = f"v1/telephony/config/people/{workspace.id}/devices"
            else:
                raise ValueError("Person or Workspace must be provided")
        else:
            raise ValueError("Cannot determine Person or Workspace from given arguments")

        if mac is None and 'ACTIVATION_CODE' in model.onboarding_method:
            # If no MAC address is provided, just generate an activation code for the device
            try:
                response = webex_api_call('post',
                                          'v1/devices/activationCode',
                                          payload=payload, params=params)
                log.debug(f"\t{response}")
            except APIError:
                return False

            # Get the ID of the device we just inserted
            device_id = response.get('id', None)
            if device_id is not None:
                device_id = device_id.replace('=', '')

            results = {
                'device_id': device_id,
                'activation_code': response['code']
            }
        else:
            payload['mac'] = mac
            if password_needed is True:
                if password is None:  # Generate a unique password
                    if isinstance(self.parent, (wxcadm.person.Person, wxcadm.person.Me)):
                        password_location = self.parent.location
                    if isinstance(self.parent, wxcadm.workspace.Workspace):
                        password_location = self.parent._parent.locations.webex_calling(single=True).id
                    if isinstance(self.parent, wxcadm.org.Org):
                        password_location = self.parent.locations.webex_calling(single=True).id
                    response = webex_api_call('POST',
                                              f'v1/telephony/config/locations/{password_location}/actions/'
                                              f'generatePassword/invoke')
                    password = response['exampleSipPassword']
                payload['password'] = password

            response = webex_api_call('post', 'v1/devices', payload=payload, params=params)
            log.debug(f"\t{response}")

            # Get the ID of the device we just inserted
            # Adding 9800s does not return any JSON, just a bool, so when that happens, we just need to go find the
            # newly added device. Unfortunately, it means an entirely different API call (4.4.1)
            if isinstance(response, bool):
                device_info = webex_api_call('get', device_info_url, params={'orgId': self.parent.org_id})
                for device in device_info['devices']:
                    if device['mac'] == mac.upper().replace(':', '').replace('-', ''):
                        new_device = Device(parent=self.parent, config=device, id=device['id'])
                        results = {
                            'device_id': new_device.id,
                            'mac': new_device.mac,
                            'device_object': new_device
                        }
            else:
                device_id = response.get('id', None).replace('=', '')

                results = {
                    'device_id': device_id,
                    'mac': response['mac'],
                    'device_object': Device(self.parent, config=response)
                }

            if data_needed is True:
                response = webex_api_call('get', f'/v1/telephony/config/devices/{device_id}',
                                          params={'orgId': self.parent.org_id})
                results['sip_auth_user'] = response['owner']['sipUserName']
                results['line_port'] = response['owner']['linePort']
                results['password'] = password
                results['sip_userpart'] = response['owner']['linePort'].split('@')[0]
                results['sip_hostpart'] = response['owner']['linePort'].split('@')[1]
                results['sip_outbound_proxy'] = response['proxy']['outboundProxy']
                results['sip_outbound_proxy_srv'] = f"_sips._tcp.{response['proxy']['outboundProxy']}"

        # Provide the Device instance in the response as well
        return results

    @property
    def supported_devices(self):
        """ The list of supported devices for the Org along with the capabilities of each """
        if self._supported_devices is None:
            self._supported_devices = SupportedDeviceList()
        return self._supported_devices

    def webex_calling(self, enabled = True) -> list:
        """ Returns a list of Device instances where Webex Calling is enabled or disabled.

        Args:
            enabled(bool, optional): Whether Webex Calling is enabled on the Device. Defaults to True.

        Returns:
            list: List of :class:`Device` instances matching the requested argument.

        """
        device_list = []
        device: Device
        for device in self.data:
            if device.workspace_location_id is not None:
                wxc_device = True
            else:
                wxc_device = False
            if wxc_device is enabled:
                device_list.append(device)
        return device_list

    def get_by_status(self, status: str) -> list:
        """ Get a list of devices by the connection status from :attr:`Device.connection_status`

        The `status` argument accepts the "raw" :attr:`Device.connection_status`:
            - "connected"
            - "disconnected"
            - "connected_with_issues"
            - "offline_expired"
            - "activating"
            - "unknown"
        It also accepts the following grouping values which are unique to **wxcadm**:
            - "online" returns devices which are known to be online
            - "offline" returns devices which are known to be offline

        Since `"activating"` and `"unknown"` devices are not known to be in a state, they can only be returned by
            passing the raw value as the status

        Args:
            status (str): See the documentation above for a list of valid values.

        Returns:
            list(Device): List of :class:`Device` instances matching the requested argument.

        """
        return_list = []
        if status.lower() == 'online':
            match_status_list = ['connected', 'connected_with_issues']
        elif status.lower() == 'offline':
            match_status_list = ['disconnected', 'offline_expired', 'offline_deep_sleep']
        else:
            match_status_list = [status]
        for device in self.data:
            if device.connection_status.lower() in match_status_list:
                return_list.append(device)
        return return_list


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class SupportedDevice:
    model: str
    display_name: str
    type: str
    manufacturer: str
    managed_by: str
    supported_for: list
    onboarding_method: list
    allow_configure_layout_enabled: bool
    number_of_line_ports: bool
    kem_support_enabled: bool
    upgrade_channel_enabled: bool
    customized_behaviors_enabled: bool
    allow_configure_ports_enabled: bool
    customizable_line_label_enabled: bool
    kem_module_count: Optional[int] = None
    kem_module_type: Optional[list] = None
    default_upgrade_channel: Optional[str] = None
    additional_primary_line_appearances_enabled: Optional[bool] = None
    basic_emergency_nomadic_enabled: Optional[bool] = None


class SupportedDeviceList(UserList):
    _endpoint = "v1/telephony/config/supportedDevices"
    _endpoint_items_key = 'devices'
    _item_endpoint = None
    _item_class = SupportedDevice

    def __init__(self):
        super().__init__()
        log.info('Collecting SupportedDeviceList')
        self.data: list = self._get_data()

    def _get_data(self) -> list:
        log.debug('_get_data() started')
        response = webex_api_call('get', self._endpoint)
        items = []
        if self._endpoint_items_key is not None:
            log.info(f"Found {len(response[self._endpoint_items_key])} items")
            for entry in response[self._endpoint_items_key]:
                items.append(self._item_class.from_dict(entry))
        else:
            log.info(f"Found {len(response)} items")
            for entry in response:
                items.append(self._item_class(**entry))
        return items

    def get(self, model_name: str):
        """ Get a :class:`SupportedDevice` of list of :class:`SupportedDevices`s matching given model name
        
        Matches will match on a partial model name. For example ``model='Cisco 8841'`` will match a
        :class:`SupportedDevice` with model 'DMS Cisco 8841'.

        Args:
            model_name (str): The name or partial model name. Not case-sensitive.

        Returns:
            SupportedDevice: If only a single entry matches, that Device will be returned. If multiple Devices match,
            a list of devices will be returned.

        """
        supported_devices = []
        for dev in self.data:
            if model_name.upper() in dev.model.upper():
                supported_devices.append(dev)
        if len(supported_devices) == 1:
            return supported_devices[0]
        else:
            return supported_devices



from __future__ import annotations
from typing import Optional
import requests
from .common import *
import wxcadm
from wxcadm import log
from datetime import datetime, timedelta
from io import BytesIO
from zipfile import ZipFile
from collections import UserList


class Report:
    def __init__(self, org: wxcadm.Org, id: Optional[str] = None, config: Optional[dict] = None):
        self.org: wxcadm.Org = org
        if id is None and config is None:
            raise ValueError('id or config is required for Report init')
        if id is not None:
            log.debug(f'Getting Report details: {id}')
            config = webex_api_call('get', f'v1/reports/{id}')[0]
            log.debug(config)
        self.id: str = config.get('Id', None)
        """ The Report ID """
        self.title: str = config.get('title', '')
        """ The title of the Report """
        self.service: str = config.get('service', '')
        """ The service to which the Report belongs """
        self.start_date: str = config.get('startDate', '')
        """ The Report period start date """
        self.end_date: str = config.get('endDate', '')
        """ The Report period end date """
        self.site_list: str = config.get('siteList', '')
        """ For Webex Meetings reports, the list of Sites """
        self.created: str = config.get('created', '')
        """ The date the report was created """
        self.created_by: str = config.get('createdBy', '')
        """ The person who created the Report """
        self.scheduled_from: str = config.get('scheduledFrom', '')
        """ Whether the Report was scheduled from Control Hub or API """
        self._status: str = config.get('status', 'unknown')
        self.download_url: str = config.get('downloadURL', None)

    def refresh(self):
        """ Refresh the Report

        .. note::
        There is no need to call refresh() before calling status(), because a full refresh will be done when the status
        is updated.

        Returns:
            Report: An updated :class:`Report` instance

        """
        log.debug(f'Refreshing Report {self.id}')
        config = webex_api_call('get', f'v1/reports/{self.id}')[0]
        log.debug(config)
        self.id = config.get('Id', None)
        self.title = config.get('title', '')
        self.service = config.get('service', '')
        self.start_date = config.get('startDate', '')
        self.end_date = config.get('endDate', '')
        self.site_list = config.get('siteList', '')
        self.created = config.get('created', '')
        self.created_by = config.get('createdBy', '')
        self.scheduled_from = config.get('scheduledFrom', '')
        self._status = config.get('status', 'unknown')
        self.download_url = config.get('downloadURL', None)

    @property
    def status(self) -> str:
        """ The status of the report """
        self.refresh()
        return self._status

    def get_report_lines(self) -> list:
        """ Get the lines of the Report as a list.

        .. note::

            No character processing is done by **wxcadm** other than splitting each line into an index. You will have to
            parse the line and remove extraneous characters in most cases.


        Returns:
            list: The lines of the Report

        """
        log.info(f'Getting Report lines: {self.id}')
        self.refresh()
        log.debug(f'Download URL: {self.download_url}')
        r = requests.get(self.download_url, headers=self.org._headers)
        log.debug(f'Response Headers: {r.headers}')
        if r.ok:
            if r.headers.get('content-type', '') in ['application/zip', 'application/octet-stream']:
                report_lines = []
                myzip = ZipFile(BytesIO(r.content))
                filename = myzip.namelist()[0]
                for line in myzip.open(filename).readlines():
                    report_lines.append(line.decode('utf-8'))
                return report_lines
            else:
                report_lines = r.text.replace('\xef\xbb\xbf', '')
                return report_lines.split('\n')
        else:
            return []


class ReportTemplate:
    def __init__(self, config: dict):
        self.id: str = config.get('Id', None)
        """ The Template ID """
        self.title: str = config.get('title', '')
        """ The title of the Report """
        self.service: str = config.get('service', '')
        """ Which Webex service the Template applies to """
        self.max_days: int = config.get('maxDays', 0)
        """ The maximum number of days that can be included in the Report """
        self.identifier: str = config.get('identifier', '')
        """ Any identifiers that can be passed to the Template """
        self.validations: list = config.get('validations', [])
        """ A list of validations which are run against the Template when a Report is requested """


class ReportList(UserList):
    """ The Reports class provides an interface to all reports available via the Webex API """
    def __init__(self, org: wxcadm.Org):
        log.info("Initializing empty DECTNetworkList")
        super().__init__()
        self.org = org
        self.data = self._get_data()
        self._templates: Optional[list] = None

    def _get_data(self) -> list:
        data = []
        response = webex_api_call('get', 'v1/reports')
        for item in response:
            data.append(Report(self.org, config=item))
        return data

    @property
    def templates(self):
        """ The Report Templates available

        Returns:
            list: A list of :class:`ReportTemplate` instances

        """
        log.info('Getting reports templates')
        if self._templates is None:
            self._templates = []
            for item in webex_api_call('get', '/v1/report/templates'):
                self._templates.append(ReportTemplate(item))
        return self._templates

    def get_template(self, id: int) -> Optional[ReportTemplate]:
        """ Get a :class:`ReportTemplate` by the Template ID

        Args:
            id (str): The Template ID to find. Normally these look like ints, but Webex expects str

        Returns:
            ReportTemplate: The :class:`ReportTemplate` matching the give ID. None is returned if no match is found.

        """
        for templ in self.templates:
            if templ.id == id:
                return templ
        return None

    def list_reports(self):
        """ List the reports available to the authorized user via the Webex API

        .. note::

            Webex isolates reports created via the API from reports created within Control Hub. The reports listed
            by this method will show only the reports created using the API.

        Returns:
            list[dict]: A list of all reports as dicts

        """
        return self.data

    def create_report(self, template: ReportTemplate,
                      start_date: Optional[str] = None,
                      end_date: Optional[str] = None,
                      site_list: Optional[list] = None):
        """ Create a new report

        The arguments vary based on the Template used. By looking at the `templates` list, you can see the requirements
        in the `validations` key.

        Args:
            template (ReportTemplate): The :class:`ReportTemplate` to use for the report
            start_date (str): The first date to include in the report
            end_date (str): The last date to include in the report
            site_list (list[str], optional): A list of Webex meeting sites

        Returns:
            Report: The Report that was created

        """
        log.info(f'Creating report with Template ID: {template.id}')
        log.debug(f'\tStart: {start_date}, End: {end_date}')
        log.debug(f'\tSite List: {site_list}')
        payload = {'templateId': template.id}
        if start_date is not None:
            payload['startDate'] = start_date
        if end_date is not None:
            payload['endDate'] = end_date
        if site_list is not None:
            payload['siteList'] = site_list

        response = webex_api_call('post', '/v1/reports', payload=payload)
        log.debug(f'API response: {response}')
        report_id = response['items']['Id']
        log.info(f'Report ID: {report_id}')
        this_report = Report(self.org, id=report_id)
        self.data.append(Report)
        return this_report

    def delete_report(self, report_id: str) -> bool:
        """ Delete a specified report

        Args:
            report_id (str): The ID of the report to delete

        Returns:
            bool: True if deletion was successful, False otherwise
        """
        log.info(f'Deleting report with ID {report_id}')
        try:
            response = webex_api_call('delete', f'/v1/reports/{report_id}')
            if response.status_code == 204:
                log.info(f'Successfully deleted report with ID: {report_id}')
                return True
            else:
                log.error(f'Failed to delete report with ID: {report_id}, Status Code: {response.status_code}')
                return False
        except Exception as e:
            log.error(f'Exception occurred while deleting report: {e}')
            return False

    def report_status(self, report_id: str):
        """ The text status of the provided report_id

        Args:
            report_id (str): The Report ID

        Returns:
            str: The status of the report

        """
        for item in self.data:
            if item.id == report_id:
                return item.status
        return None

    def get_report_lines(self, report_id: str):
        """ Download and return the lines from a report

        This method returns the full report in a list, with each line as an entry.

        Args:
            report_id (str): The Report ID

        Returns:
            list: The report lines

        """
        for report in self.data:
            if report.id == report_id:
                return report.get_report_lines()
        return None

    def cdr_report(self, start: Optional[str] = None, end: Optional[str] = None, days: Optional[int] = None):
        """ Create a CDR report

        This method serves as a shortcut to :py:meth:`create_report()` for a Calling Detailed Call History report. It
        finds the correct template and creates the report using the given start and end dates, or the ``days``
        argument can be provided to report the last X days.

        Args:
            start (str, optional): The first date to include in the report (YYYY-MM-DD)
            end (str, optional): The last date to include in the report (YYYY-MM-DD)
            days (int, optional): The number of days to report on, ending on yesterday's date

        Returns:
            str: The Report ID

        """
        if start is None and days is None:
            raise ValueError("You must provide either a start date or the number of days")
        if start is not None and days is not None:
            raise ValueError("You must not specify a start date and number of days")
        if days is not None:
            yesterday = datetime.now() - timedelta(1)
            end = datetime.strftime(yesterday, "%Y-%m-%d")
            start_date = yesterday - timedelta(days)
            start = datetime.strftime(start_date, "%Y-%m-%d")

        # Find the CDR Report Template
        for template in self.templates:
            if template.title == 'Calling Detailed Call History':
                cdr_template = template
                break

        new_report = self.create_report(template=cdr_template,
                                       start_date=start,
                                       end_date=end)
        return new_report



from __future__ import annotations

from typing import Union
from dataclasses import dataclass, field
from collections import UserList

import wxcadm.org
from wxcadm import log
from .common import *


class WebexApplications(UserList):
    def __init__(self, parent: wxcadm.Org):
        super().__init__()
        log.debug("Initializing WebexApps instance")
        self.parent: wxcadm.Org = parent
        self.data: list = self._get_applications()

    def _get_applications(self):
        app_list = []
        apps = webex_api_call("get", "v1/applications", params={"orgId": self.parent.id})
        for app in apps:
            this_app = WebexApplication(parent=self.parent, **app)
            app_list.append(this_app)
        return app_list

    @property
    def org_apps(self):
        """ List of :py:class:`WebexApplication` instances for apps owned by this Org """
        org_apps = []
        for app in self.data:
            if app.orgId == self.parent.id:
                org_apps.append(app)
        return org_apps

    def get_app_by_name(self, name: str) -> Union[WebexApplication, list, None]:
        """ Get a :py:class:`WebexApplication` instance by application name

        .. note::
            It isn't "against the rules" for two apps to have the same name. If only one app is found, just the
            :py:class:`WebexApplication` instance is returned. If more than one app is found, a list of instances will
            be returned

        Args:
             name (str): The name of the :py:class:`WebexApplication` to search for

        Returns:
            WebexApplication: The application instance

        """
        found_apps = []
        for app in self.data:
            if app.name == name:
                found_apps.append(app)
        if len(found_apps) == 0:
            return None
        elif len(found_apps) == 1:
            return found_apps[0]
        else:
            return found_apps

    def get_app_by_id(self, id: str) -> WebexApplication:
        """ Get a :py:class:`WebexApplication` instance by application ID

        Args:
            id (str): The ID of the :py:class:`WebexApplication` to get

        Returns:
            WebexApplication: The application instance. None is returned if no match is found.

        """
        # First check the known apps to see if we already have it
        for app in self.data:
            if app.id == id:
                return app
        # If it wasn't found, call the API to get the app and build an instance for it
        app = webex_api_call('get', f'/v1/applications/{id}')
        this_app = WebexApplication(parent=self.parent, **app)
        self.data.append(this_app)
        return this_app

    def add_service_application(self, name: str,
                                contact_email: str,
                                scopes: list,
                                logo: str = "https://pngimg.com/uploads/hacker/hacker_PNG6.png"):
        """ Add a Service App to the Org

        .. note::

            Ensure you record the client_secret that comes back. It will not be visible later.

        .. warning::

            The Service App capability is in Early Field Trial and may not be available for you Org.

        Args:
            name (str): The name of the application
            contact_email (str): The Contact Email to attach to the application
            scopes (list): A list of Webex scopes to assign to the application
            logo (str, optional): A URL to a logo image. If not specified, a generic phone icon is used.

        """
        payload = {'type': 'authorizedAppIssuer',
                   'name': name,
                   'contactEmail': contact_email,
                   'logo': logo,
                   'scopes': scopes}
        response = webex_api_call("post", "v1/applications", payload=payload)
        return response


@dataclass
class WebexApplication:
    parent: wxcadm.Org = field(repr=False)
    isNative: bool = field(repr=False)
    id: str
    friendlyId: str
    type: str
    name: str
    orgId: str = field(repr=False)
    isFeatured: bool = field(repr=False)
    submissionStatus: str = field(repr=False)
    createdBy: str = field(repr=False)
    created: str = field(repr=False)
    description: str = field(default='', repr=False)
    modified: str = field(default='', repr=False)
    redirectUrls: list = field(default=None, repr=False)
    scopes: list = field(default=None, repr=False)
    clientId: str = field(default='', repr=False)
    logo: str = field(default='', repr=False)
    tagLine: str = field(default='', repr=False)
    shortTagLine: str = field(default='', repr=False)
    categories: list = field(default=None, repr=False)
    videoUrl: str = field(default='', repr=False)
    contactEmail: str = field(default='', repr=False)
    contactName: str = field(default='', repr=False)
    companyName: str = field(default='', repr=False)
    companyUrl: str = field(default='', repr=False)
    supportUrl: str = field(default='', repr=False)
    privacyUrl: str = field(default='', repr=False)
    productUrl: str = field(default='', repr=False)
    applicationUrls: list = field(default=None, repr=False)
    botEmail: str = field(default='', repr=False)
    botPersonId: str = field(default='', repr=False)
    submissionDate: str = field(default='', repr=False)
    orgSubmissionStatus: str = field(default='', repr=False)
    appContext: str = field(default='', repr=False)
    tags: list = field(default=None, repr=False)
    screenshot1: str = field(default='', repr=False)
    screenshot2: str = field(default='', repr=False)
    screenshot3: str = field(default='', repr=False)
    validDomains: list = field(default=None, repr=False)
    groupId: str = field(default='', repr=False)
    version: str = field(default='', repr=False)
    meetingLayoutPreference: str = field(default='', repr=False)
    entitlements: str = field(default='', repr=False)

    def authorize(self):
        """ Authorize the application for the Organization

        Returns:
            bool: True on success, False otherwise

        """
        payload = {'orgId': self.parent.id}
        response = webex_api_call('post', f'/v1/applications/{self.id}/authorize', payload=payload)
        return response

    def delete(self):
        """ Delete the application

        Returns:

        """
        response = webex_api_call('delete', f'/v1/applications/{self.id}')
        return response

    def get_token(self, client_secret: str, target_org: wxcadm.Org | str):
        """ Get the access and refresh tokens to utilize the application in the target Org

        Args:
            client_secret (str): The Client Secret value stored when the Application was created
            target_org (str): The Base64 Org ID to obtain a token for

        Returns:
            dict: A dict of the token info, including access_token and refresh_token

        """
        if isinstance(target_org, wxcadm.org.Org):
            org_id = target_org.id
        else:
            org_id = target_org
            # TODO - Eventually we should take the Org ID in both Base64 (current) and the UUID format

        payload = {
            'clientId': self.clientId,
            'clientSecret': client_secret,
            'targetOrgId': org_id
        }
        response = webex_api_call('post', f'/v1/applications/{self.id}/token', payload=payload)
        return response

    def get_token_refresh(self, client_secret: str, refresh_token: str):
        """ Use the provided Refresh Token to obtain a new Access Token

        Args:
            client_secret (str): The Client Secret value stored when the Application was created
            refresh_token (str): The Refresh Token stored when the initial token was obtained with :py:meth:get_token()

        Returns:
            dict: A dict of the token info, including access_token and refresh_token

        """
        payload = {
            'grant_type': 'refresh_token',
            'client_id': self.clientId,
            'client_secret': client_secret,
            'refresh_token': refresh_token
        }
        response = webex_api_call('post', '/v1/access_token', payload=payload)
        return response

    def regenerate_client_secret(self):
        """ Obtain a new Client Secret if the initial value was lost or compromised

        Returns:
            str: The new Client Secret value
        """
        response = webex_api_call('delete', f'/v1/applications/{self.id}/clientSecret')
        if 'clientSecret' in response:
            return response['clientSecret']
        else:
            return False



import unittest
import os
from dotenv import load_dotenv
import wxcadm
from random import choice
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    pass


class TestTranslationPatterns(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        load_dotenv()
        cls.access_token = os.getenv("WEBEX_ACCESS_TOKEN")
        if not cls.access_token:
            print("No WEBEX_ACCESS_TOKEN found. Cannot continue.")
            exit(1)

    def setUp(self) -> None:
        self.webex = wxcadm.Webex(self.access_token)
        self.random_location = choice(self.webex.org.locations.webex_calling())
        # Enable for test debugging
        # wxcadm.console_logging()

    def test_org_tp_instances(self) -> None:
        org_tp = self.webex.org.translation_patterns
        self.assertIsInstance(org_tp, wxcadm.TranslationPatternList)
        if len(org_tp) > 0:
            self.assertIsInstance(org_tp[0], wxcadm.TranslationPattern)

    def test_org_tp_crud(self) -> None:
        with self.subTest("Create"):
            new_pattern = self.webex.org.translation_patterns.create(
                name='WXCADM Test',
                match_pattern='1234567890123456789X',
                replacement_pattern='8443876962'
            )
            self.assertIsInstance(new_pattern, wxcadm.TranslationPattern)
        with self.subTest("Retrieve"):
            self.webex.org.translation_patterns.refresh()
            pattern = self.webex.org.translation_patterns.get(id=new_pattern.id)
            self.assertIsInstance(pattern, wxcadm.TranslationPattern)
        with self.subTest("Update Name"):
            success = pattern.update(name='WXCADM Updated Test')
            self.assertTrue(success)
        with self.subTest("Update Match Pattern"):
            success = pattern.update(match_pattern='987654321098765432X')
            self.assertTrue(success)
        with self.subTest("Update Replacement Pattern"):
            success = pattern.update(replacement_pattern='18553876962')
            self.assertTrue(success)
        with self.subTest("Delete"):
            success = pattern.delete()
            self.assertTrue(success)

    def test_loc_tp_crud(self) -> None:
        with self.subTest("Create"):
            new_pattern = self.webex.org.translation_patterns.create(
                name='WXCADM Loc Test',
                match_pattern='1234567890123456789X',
                replacement_pattern='8443876962',
                location=self.random_location
            )
            self.assertIsInstance(new_pattern, wxcadm.TranslationPattern)
        with self.subTest("Retrieve"):
            self.webex.org.translation_patterns.refresh()
            pattern = self.webex.org.translation_patterns.get(id=new_pattern.id)
            self.assertIsInstance(pattern, wxcadm.TranslationPattern)
        with self.subTest("Update Name"):
            success = pattern.update(name='WXCADM Updated Loc Test')
            self.assertTrue(success)
        with self.subTest("Update Match Pattern"):
            success = pattern.update(match_pattern='987654321098765432X')
            self.assertTrue(success)
        with self.subTest("Update Replacement Pattern"):
            success = pattern.update(replacement_pattern='18553876962')
            self.assertTrue(success)
        with self.subTest("Delete"):
            success = pattern.delete()
            self.assertTrue(success)

class TestTrunks(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        load_dotenv()
        cls.access_token = os.getenv("WEBEX_ACCESS_TOKEN")
        if not cls.access_token:
            print("No WEBEX_ACCESS_TOKEN found. Cannot continue.")
            exit(1)

    def setUp(self) -> None:
        self.webex = wxcadm.Webex(self.access_token)
        self.random_location = choice(self.webex.org.locations.webex_calling())
        # Enable for test debugging
        # wxcadm.console_logging()

    def test_trunks_list(self) -> None:
        trunks = self.webex.org.call_routing.trunks
        self.assertIsInstance(trunks, wxcadm.Trunks)
        for trunk in trunks:
            self.assertIsInstance(trunk, wxcadm.Trunk)
            with self.subTest("Get Details"):
                status = trunk.status
                self.assertIsInstance(status, str)
                self.assertIsInstance(trunk.outbound_proxy, (wxcadm.OutboundProxy, None))


if __name__ == '__main__':
    unittest.main()


from __future__ import annotations
from typing import Optional
from datetime import datetime, timedelta
from collections import UserList
import requests

from .common import *
import wxcadm
from wxcadm import log


class AuditEvent:
    def __init__(self, config: dict):
        self.id: str = config.get('id', '')
        """ The ID of the Admin Event """
        self.actor_id: str = config.get('actorId', '')
        """ The ID of the actor making the change """
        self.actor_name: str = config['data'].get('actorName', '')
        """ The name of the actor making the change """
        self.actor_email: str = config['data'].get('actorEmail', '')
        """ The email address of the actor making the change """
        self.actor_org_name: str = config['data'].get('actorOrgName', '')
        """ The Org name of the actor making the change """
        self.actor_org_id: str = config.get('actorOrgId', '')
        """ The Org ID of the actor making the change """
        self.actor_user_agent: str = config['data'].get('actorUserAgent', '')
        """ The User Agent type of the actor making the change """
        self.admin_roles: list[str] = config['data'].get('adminRoles', [])
        """ The admin roles of the actor making the change """
        self.actor_ip: str = config['data'].get('actorIp', '')
        """ The IP address of the actor making the change """
        self.target_type: str = config['data'].get('targetType', '')
        """ The type of target being changed """
        self.target_id: str = config['data'].get('targetId', '')
        """ The identified of the target being changed """
        self.target_name: str = config['data'].get('targetName', '')
        """ The name of the target being changed """
        self.target_org_name: str = config['data'].get('targetOrgName', '')
        """ The Org name of the target being changed """
        self.target_org_id: str = config['data'].get('targetOrgId', '')
        """ The Org ID of the target being changed """
        self.description: str = config['data'].get('eventDescription', '')
        """ The description of the change """
        self.tracking_id: str = config['data'].get('trackingId', '')
        """ A unique tracking ID for the change """
        self.category: str = config['data'].get('eventCategory', '')
        """ The category of the change """
        self.text: str = config['data'].get('actionText', '')
        """ The full text of the change details """
        self.timestamp: str = config.get('created', '')
        """ The timestamp when the change was made """


class AuditEventList(UserList):
    def __init__(self, parent: wxcadm.Org, start: str, end: str):
        log.info("AuditEventList instance created")
        super().__init__()
        self.parent = parent
        self.start = start
        self.end = end
        self.data = self._get_data()

    def _get_data(self):
        data = []
        response = webex_api_call(
            "get",
            f"v1/adminAudit/events",
            params={'orgId': self.parent.org_id, 'from': self.start, 'to': self.end}
        )
        for entry in response:
            data.append(AuditEvent(entry))
        return data




from __future__ import annotations

import logging
# Set up NullHandler for logging
log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
log.addHandler(logging.NullHandler())

from .webex import Webex
from .models import *
from .org import Org
from .xsi import XSIEvents, Call, XSI, XSICallQueue
from .redsky import RedSky
from .meraki import Meraki
from .cdr import CallDetailRecords
from .exceptions import *
from .common import *
from .wholesale import Wholesale
from .location_features import *
from .announcements import *
from .applications import *
from .auto_attendant import *
from .call_queue import *
from .call_routing import *
from .calls import *
from .dect import *
from .device import *
from .hunt_group import *
from .jobs import *
from .location import *
from .meraki import *
from .monitoring import *
from .number import *
from .person import *
from .pickup_group import *
from .recording import *
from .redsky import *
from .reports import *
from .virtual_line import *
from .webhooks import *
from .workspace import *
from .pstn import *



from __future__ import annotations

from typing import Optional, TYPE_CHECKING
if TYPE_CHECKING:
    from .person import Person
from collections import UserList

import wxcadm.org
from wxcadm import log
from .common import *
from .location import Location


class NumberManagementJob:
    """ The class for Number Moves.

    This class is never instantiated directly. It is created with the :class:`NumberMoveJobList`

    """

    def __init__(self, parent: wxcadm.Org, id: str, details: Optional[dict] = None):
        self.parent = parent
        self.id = id
        """ The Number Move Job ID """
        self.details: dict
        """ The full details of the Number Move Job from Webex """
        if details is None:
            self.details = self._get_details()
        else:
            self.details = details

    def _get_details(self):
        response = webex_api_call('get', f'v1/telephony/config/jobs/numbers/manageNumbers/{self.id}',
                                  params={'orgId': self.parent.id})
        return response

    @property
    def completed(self) -> bool:
        """ Whether the job has completed execution

        Note that this does not indicate that the job was successful. The :attr:`success` attribute should be used to
        determine whether the job was successful or not

        """
        self.details = self._get_details()
        if self.details['latestExecutionStatus'] == 'COMPLETED':
            return True
        else:
            return False

    @property
    def success(self) -> bool:
        """ Whether the job was successful

        This property will return False even if the job is not complete. It should really be called only after
        :attr:`completed` is True

        """
        self.details = self._get_details()
        if int(self.details['counts']['numbersFailed']) > 0:
            return False

        total_numbers = int(self.details['counts']['totalNumbers'])
        success_total = int(self.details['counts']['numbersDeleted']) + int(self.details['counts']['numbersMoved'])

        if success_total == total_numbers:
            return True
        else:
            return False


class NumberManagementJobList(UserList):
    _endpoint = "v1/telephony/config/jobs/numbers/manageNumbers"
    _endpoint_items_key = "items"
    _item_endpoint = "v1/telephony/config/jobs/numbers/manageNumbers/{item_id}"
    _item_class = NumberManagementJob

    def __init__(self, parent: wxcadm.Org):
        super().__init__()
        log.debug("Initializing NumberMoveJobList")
        self.parent: wxcadm.Org = parent
        self.data: list = self._get_data()

    def _get_data(self) -> list:
        log.debug("_get_data() started")
        params = {}

        if isinstance(self.parent, wxcadm.Org):
            log.debug(f"Using Org ID {self.parent.id} as Data filter")
            params['orgId'] = self.parent.id
            params['max'] = 1000
        else:
            log.warn("Parent class is not Org so all items will be returned")
        response = webex_api_call('get', self._endpoint, params=params)
        log.info(f"Found {len(response)} items")

        items = []
        for entry in response:
            items.append(self._item_class(parent=self.parent, id=entry['id'], details=entry))
        return items

    def refresh(self):
        """ Refresh the list of instances from Webex

        Returns:
            bool: True on success, False otherwise.

        """
        self.data = self._get_data()
        return True

    def get(self, id: str):
        """ Get the instance associated with a given Job ID

        Args:
            id (str, optional): The Number Management Job ID to find

        Returns:
            NumberManagementJob: The Number Management Job instance correlating with the given ID

        """
        for item in self.data:
            if item.id == id:
                return item
        return None

    def create(self,
               job_type: str,
               target_location: wxcadm.Location | str,
               numbers: list
               ):
        """ Create a Number Management Job

        .. note::

            At this time, only the "move" value is accepted for the `job_type` argument, because Webex only supports
            move operations via the API. Also, the API currently limits a Number Move to a single number per job. The
            API schema is written to support more, so it may be supported at some point, but for now, the `numbers` list
            must contain only a single entry.

        Args:
            job_type (str): Only "move" is supported currently
            target_location (Location | str): A :class:`Location` instance or a string representing the Location ID to
            move the numbers to
            numbers (list[str]): A list of numbers, as strings, to move to the target Location

        Returns:
            NumberManagementJob: The created NumberManagementJob

        """
        # Determine if the numbers are valid
        numbers_to_move = []
        # Get a copy of the Org numbers for processing
        org_numbers = self.parent.numbers
        for number in numbers:
            number_found = False
            for org_number in org_numbers:
                if org_number.phone_number is not None:
                    if number in org_number.phone_number:
                        number_found = True
                        numbers_to_move.append(org_number)
            if number_found is False:
                raise KeyError(f"{number} was not found in the Org")

        if isinstance(target_location, wxcadm.Location):
            target_location_id = target_location.id
        else:
            target_location_id = target_location

        # Build the payload for the move job
        payload = {'operation': 'MOVE', 'targetLocationId': target_location_id, 'numberList': []}
        for number in numbers_to_move:
            payload['numberList'].append({'locationId': number.location.id, 'numbers': [number.phone_number]})

        response = wxcadm.webex_api_call('post', 'v1/telephony/config/jobs/numbers/manageNumbers',
                                         params={'orgId': self.parent.id},
                                         payload=payload)
        job_id = response['id']
        return NumberManagementJob(parent=self.parent, id=job_id)


class UserMoveJob:
    """ The class for User Location Moves

    This class is never instantiated directly. It is created with the :class:`UserMoveJobList`

    """

    def __init__(self, parent: wxcadm.Org, id: str, details: Optional[dict] = None):
        self.parent = parent
        self.id = id
        """ The User Move Job ID """
        self.details: dict
        """ The full details of the User Move Job from Webex """
        if details is None:
            self.details = self._get_details()
        else:
            self.details = details

    def _get_details(self):
        response = webex_api_call('get', f'v1/telephony/config/jobs/person/moveLocation/{self.id}',
                                  params={'orgId': self.parent.id})
        return response

    @property
    def completed(self) -> bool:
        """ Whether the job has completed execution

        Note that this does not indicate that the job was successful. The :attr:`success` attribute should be used to
        determine whether the job was successful or not

        """
        self.details = self._get_details()
        if self.details['latestExecutionStatus'] == 'COMPLETED':
            return True
        else:
            return False

    @property
    def success(self) -> bool:
        """ Whether the job was successful

        This property will return False even if the job is not complete. It should really be called only after
        :attr:`completed` is True

        """
        self.details = self._get_details()
        if int(self.details['counts']['failed']) > 0:
            return False

        total_numbers = int(self.details['counts']['totalMoves'])
        success_total = int(self.details['counts']['moved'])

        if success_total == total_numbers:
            return True
        else:
            return False


class UserMoveJobList(UserList):
    _endpoint = "v1/telephony/config/jobs/person/moveLocation"
    _endpoint_items_key = "items"
    _item_endpoint = "v1/telephony/config/jobs/person/moveLocation/{item_id}"
    _item_class = UserMoveJob

    def __init__(self, parent: wxcadm.Org):
        super().__init__()
        log.debug("Initializing UserMoveJobList")
        self.parent: wxcadm.Org = parent
        self.data: list = self._get_data()

    def _get_data(self) -> list:
        log.debug("_get_data() started")
        params = {}

        if isinstance(self.parent, wxcadm.Org):
            log.debug(f"Using Org ID {self.parent.id} as Data filter")
            params['orgId'] = self.parent.id
            params['max'] = 1000
        else:
            log.warn("Parent class is not Org so all items will be returned")
        response = webex_api_call('get', self._endpoint, params=params)
        log.info(f"Found {len(response)} items")

        items = []
        for entry in response:
            items.append(self._item_class(parent=self.parent, id=entry['id'], details=entry))
        return items

    def refresh(self):
        """ Refresh the list of instances from Webex

        Returns:
            bool: True on success, False otherwise.

        """
        self.data = self._get_data()
        return True

    def get(self, id: str):
        """ Get the instance associated with a given Job ID

        Args:
            id (str, optional): The Number Management Job ID to find

        Returns:
            NumberManagementJob: The Number Management Job instance correlating with the given ID

        """
        for item in self.data:
            if item.id == id:
                return item
        return None

    def create(self,
               target_location: wxcadm.Location | str,
               people: list[wxcadm.person.Person],
               validate_only: Optional[bool] = False
               ):
        """ Create a User Move Job

        .. note::

            At this time, the API currently limits a User Move to a single user per job. The
            API schema is written to support more, so it may be supported at some point, but for now, the `users` list
            must contain only a single entry.

        Args:
            target_location (Location | str): A :class:`Location` instance or a string representing the Location ID to
                move the numbers to

            people (list[Person]): A list of :class:`Person` instances to move to the target Location

            validate_only (bool, optional): Whether to only perform the validation of the User Move, not actually move
            the users. Defaults to False. False wll perform both a validation and create a User Move Job

        Returns:
            UserMoveValidationResults: The Results of the validation for the User Move
            UserMoveJob: The created UserMoveJob (if `validate_only` is not True and the validation succeeds). If
            validation fails, None will be returned.

        """
        log.info(f"Creating User Move Job")
        if isinstance(target_location, wxcadm.Location):
            target_location_id = target_location.id
        else:
            target_location_id = target_location

        # If we create a job, we'll store it. Otherwise, we'll return None
        user_move_job = None

        # Build the payload for the move job validation
        payload = {'usersList': []}
        for person in people:
            person_payload = {'locationId': target_location_id,
                              'validate': True,
                              'users': [{'userId': person.id, 'extension': person.extension}]}
            payload['usersList'].append(person_payload)

        validation_results = UserMoveValidationResults()
        try:
            response = wxcadm.webex_api_call('post', 'v1/telephony/config/jobs/person/moveLocation',
                                             params={'orgId': self.parent.id},
                                             payload=payload)
        except wxcadm.exceptions.APIError as e:
            log.debug(f"Webex returned an error to Number Move Validation: {e}")

            error_code = e.args[0]['errors'][0]['code']
            error_message = e.args[0]['errors'][0]['message']
            validation_results.add_message('error', error_code, error_message)
        else:
            if 'errors' in response['response']['usersList'][0].keys():
                for error in response['response']['usersList'][0]['errors']:
                    validation_results.add_message('error', error['code'], error['message'])
            if 'impacts' in response['response']['usersList'][0].keys():
                for impact in response['response']['usersList'][0]['impacts']:
                    validation_results.add_message('impact', impact['code'], impact['message'])

        # Create the job if the validation worked and if we weren't asked to only validate
        if validation_results.passed is True and validate_only is not True:
            log.debug(f"Validation passed and Move requested. Creating UserMoveJob call.")
            payload['usersList'][0]['validate'] = False
            try:
                response = wxcadm.webex_api_call('post', 'v1/telephony/config/jobs/person/moveLocation',
                                                 params={'orgId': self.parent.id},
                                                 payload=payload)
            except wxcadm.exceptions.APIError as e:
                log.debug(f"Webex returned an error to Number Move Validation: {str(e)}")
            else:
                log.debug(f"API response: {response}")
                job_id = response['response']['jobDetails']['id']
                log.debug(f"Received Job ID {job_id}")
                self.refresh()
                user_move_job = UserMoveJob(parent=self.parent, id=job_id)

        return validation_results, user_move_job


class UserMoveValidationResults:
    def __init__(self):
        self.messages = []
        """ The messages and their severity. A severity of `error` will prevent the Use Move from being completed. """

    def add_message(self, severity: str, code: Optional[str] = None, message: Optional[str] = None):
        """ Add a message to the Validation Results

        Args:
            severity (str): Either 'error' or 'impact'
            code (str, optional): The code assigned to the error.
            message (str, optional): The validation message

        Returns:
            bool: Always True

        """
        self.messages.append({'severity': severity, 'code': code, 'message': message})
        return True

    @property
    def passed(self) -> bool:
        """ Whether the validation passed with no errors and the User Move would be successful """
        for message in self.messages:
            if message['severity'].lower() == 'error':
                return False
        return True


class RebuildPhonesJob:
    """ The class for Rebuild Phones Jobs

    This class is never instantiated directly. It is created with the :class:`RebuildPhonesJobList`

    """

    def __init__(self, parent: wxcadm.Org, id: str, details: Optional[dict] = None):
        self.parent = parent
        self.id = id
        """ The Rebuild Phones Job ID """
        self.details: dict
        """ The full details of the Rebuild Phones Job from Webex """
        if details is None:
            self.details = self._get_details()
        else:
            self.details = details

    def _get_details(self):
        response = webex_api_call('get', f'v1/telephony/config/jobs/devices/rebuildPhones/{self.id}',
                                  params={'orgId': self.parent.id})
        return response

    @property
    def completed(self) -> bool:
        """ Whether the job has completed execution

        Note that this does not indicate that the job was successful. The :attr:`success` attribute should be used to
        determine whether the job was successful or not

        """
        self.details = self._get_details()
        if self.details['latestExecutionStatus'] == 'COMPLETED':
            return True
        else:
            return False

    @property
    def device_count(self) -> int:
        """ The count of devices that were included in this Rebuild Phones Job

        Returns:
            int: The count of phones

        """
        return self.details['deviceCount']


class RebuildPhonesJobList(UserList):
    _endpoint = "v1/telephony/config/jobs/devices/rebuildPhones"
    _endpoint_items_key = "items"
    _item_endpoint = "v1/telephony/config/jobs/devices/rebuildPhones/{item_id}"
    _item_class = RebuildPhonesJob

    def __init__(self, parent: wxcadm.Org):
        super().__init__()
        log.debug("Initializing RebuildPhonesJobList")
        self.parent: wxcadm.Org = parent
        self.data: list = self._get_data()

    def _get_data(self) -> list:
        log.debug("_get_data() started")
        params = {}

        if isinstance(self.parent, wxcadm.Org):
            log.debug(f"Using Org ID {self.parent.id} as Data filter")
            params['orgId'] = self.parent.id
            params['max'] = 1000
        else:
            log.warn("Parent class is not Org so all items will be returned")
        response = webex_api_call('get', self._endpoint, params=params)
        log.info(f"Found {len(response)} items")

        items = []
        for entry in response:
            items.append(self._item_class(parent=self.parent, id=entry['id'], details=entry))
        return items

    def refresh(self):
        """ Refresh the list of instances from Webex

        Returns:
            bool: True on success, False otherwise.

        """
        self.data = self._get_data()
        return True

    def get(self, id: str):
        """ Get the instance associated with a given Job ID

        Args:
            id (str): The Number Management Job ID to find

        Returns:
            RebuildPhonesJob: The Number Management Job instance correlating with the given ID

        """
        for item in self.data:
            if item.id == id:
                return item
        return None

    def create(self,
               location: wxcadm.Location | str
               ):
        """ Create a Rebuild Phones Job

        .. note::

            At this time, the API only supports rebuild at a Location level.

        Args:
            location (Location | str): A :class:`Location` instance or a string representing the Location ID for the
            rebuild

        Returns:
            RebuildPhonesJob: The created :class:`RebuildPhonesJob`

        """
        log.info(f"Creating Rebuild Phones Job")
        if isinstance(location, wxcadm.Location):
            target_location_id = location.id
        else:
            target_location_id = location

        # If we create a job, we'll store it. Otherwise, we'll return None
        job = None

        # Build the payload for the move job validation
        payload = {'locationId': target_location_id}
        response = wxcadm.webex_api_call('post', 'v1/telephony/config/jobs/devices/rebuildPhones',
                                         params={'orgId': self.parent.id},
                                         payload=payload)
        log.debug(f"API response: {response}")
        job_id = response['id']
        log.debug(f"Received Job ID {job_id}")
        self.refresh()
        job = RebuildPhonesJob(parent=self.parent, id=job_id, details=response)

        return job




import unittest
import os
from dotenv import load_dotenv
import wxcadm
from random import choice

class TestRedSkyConfig(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        load_dotenv()
        cls.redsky_user = os.getenv("REDSKY_USER")
        cls.redsky_pass = os.getenv("REDSKY_PASS")
        if not cls.redsky_pass or not cls.redsky_user:
            print("No REDSKY credentials found. Cannot continue.")
            exit(1)

    def setUp(self) -> None:
        self.redsky = wxcadm.RedSky(self.redsky_user, self.redsky_pass)

    def test_get_all_locations(self):
        locations = self.redsky.get_all_locations()
        self.assertIsInstance(locations, dict)
        self.assertIsInstance(locations['corporate'], list)
        self.assertIsInstance(locations['personal'], list)

    def test_buildings(self):
        buildings = self.redsky.buildings
        self.assertIsInstance(buildings, list)

    def test_held_devices(self):
        devices = self.redsky.held_devices
        self.assertIsInstance(devices, list)

    def test_phones_without_location(self):
        devices = self.redsky.phones_without_location()
        self.assertIsInstance(devices, list)

    def test_get_mac_discovery(self):
        mac = self.redsky.get_mac_discovery()
        self.assertIsInstance(mac, list)

    def test_get_lldp_discovery(self):
        lldp = self.redsky.get_lldp_discovery()
        self.assertIsInstance(lldp, list)

    def test_get_bssid_discovery(self):
        bssid = self.redsky.get_bssid_discovery()
        self.assertIsInstance(bssid, list)

    def test_get_ip_range_discovery(self):
        ip = self.redsky.get_ip_range_discovery()
        self.assertIsInstance(ip, list)


class TestRedSkyUsers(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        load_dotenv()
        cls.redsky_user = os.getenv("REDSKY_USER")
        cls.redsky_pass = os.getenv("REDSKY_PASS")
        if not cls.redsky_pass or not cls.redsky_user:
            print("No REDSKY credentials found. Cannot continue.")
            exit(1)

    def setUp(self) -> None:
        self.redsky = wxcadm.RedSky(self.redsky_user, self.redsky_pass)

    def test_get_users(self):
        userlist = self.redsky.users
        self.assertIsInstance(userlist, wxcadm.redsky.RedSkyUsers)


class TestRedSkyBuilding(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        load_dotenv()
        cls.redsky_user = os.getenv("REDSKY_USER")
        cls.redsky_pass = os.getenv("REDSKY_PASS")
        if not cls.redsky_pass or not cls.redsky_user:
            print("No REDSKY credentials found. Cannot continue.")
            exit(1)

    def setUp(self) -> None:
        self.redsky = wxcadm.RedSky(self.redsky_user, self.redsky_pass)

    def test_get_building(self):
        building = choice(self.redsky.buildings)
        self.assertIsInstance(building, wxcadm.redsky.RedSkyBuilding)
        with self.subTest("Building Locations"):
            locations = building.locations
            self.assertIsInstance(locations, list)
            self.assertIsInstance(locations[0], wxcadm.redsky.RedSkyLocation)



import unittest
import os
from dotenv import load_dotenv
import wxcadm
from random import choice


class TestJobsList(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        load_dotenv()
        cls.access_token = os.getenv("WEBEX_ACCESS_TOKEN")
        if not cls.access_token:
            print("No WEBEX_ACCESS_TOKEN found. Cannot continue.")
            exit(1)

    def setUp(self) -> None:
        self.webex = wxcadm.Webex(self.access_token)
        self.random_location = choice(self.webex.org.locations.webex_calling())

    def test_jobs_list(self):
        jobs = self.webex.org.number_management_jobs
        self.assertIsInstance(jobs, wxcadm.jobs.NumberManagementJobList)

    def test_job_details(self):
        jobs = self.webex.org.number_management_jobs
        with self.subTest("NumberManagementJob instance"):
            job = choice(jobs)
            self.assertIsInstance(job, wxcadm.jobs.NumberManagementJob)

        with self.subTest("Job Details"):
            details = choice(jobs).details
            self.assertIsInstance(details, dict)

        with self.subTest("Job Completed"):
            self.assertIsInstance(choice(jobs).completed, bool)

        with self.subTest("Job Success"):
            self.assertIsInstance(choice(jobs).success, bool)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

from typing import Optional
from dataclasses import dataclass, field
from dataclasses_json import dataclass_json, LetterCase
from collections import UserList
from datetime import datetime

import wxcadm.location
from .realtime import RealtimeClass
from wxcadm import log
from .common import *


@dataclass
class PagingGroup:
    parent: wxcadm.Location
    id: str
    name: str
    spark_id: str = field(init=False, repr=False)
    config: dict = field(init=False, repr=False)

    def __post_init__(self):
        self.spark_id = decode_spark_id(self.id)
        self.refresh_config()

    def refresh_config(self):
        """ Pull a fresh copy of the Paging Group config from Webex, in case it has changed. """
        api_resp: dict = webex_api_call("get", f"v1/telephony/config/locations/{self.parent.id}/paging/{self.id}")
        self.config = api_resp


@dataclass
class LocationSchedule:
    """ The class for Location Schedules within a :class:`Location`

    When the instance is initialized, it will fetch the configuration from Webex.

    .. note::
            LocationSchedule instances are normally not instantiated manually and are done automatically with the
            :attr:`Location.schedules` property.

    Args:
        parent (Location): The `Location` instance to which the LocationSchedule is assigned.
        id (str): The Webex ID of the LocationSchedule
        name (str): The name of the LocationSchedule
        type (str): The type of LocationSchedule, either 'businessHours' or 'Holidays'

    """
    parent: wxcadm.Location
    """ The `Location` instance to which the LocationSchedule is assigned """
    id: str
    """ The Webex ID of the LocationSchedule """
    name: str
    """ The name of the LocationSchedule """
    type: str
    """ The type of LocationSchedule, either 'businessHours' or 'Holidays' """
    config: dict = field(init=False, repr=False)
    """ The configuration returned by Webex """

    def __post_init__(self):
        self.refresh_config()

    def refresh_config(self):
        """ Pull a fresh copy of the schedule configuration from Webex, in case it has changed. """
        api_resp = webex_api_call("get", f"v1/telephony/config/locations/{self.parent.id}/schedules/"
                                         f"{self.type}/{self.id}")
        self.config = api_resp

    def add_holiday(self, name: str, date: str, recur: bool = False, recurrence: Optional[dict] = None):
        """ Add a new all-day event to a Holidays schedule

        This method provides a quicker way to add an all-day event to a Holidays schedule than using
        :py:meth:`add_event()`. When used on a businessHours schedule, a TypeError will be raised.

        If :attr:`recur` is True and the :attr:`recurrence` dict is not present, the method will set the holiday to
        repeat yearly on the same month and day, which works for holidays like Christmas that fall on the same date.
        If you need to add a holiday, such as Thanksgiving, that recurs on the 4th Thursday of November,
        the :attr:`recurrence` dictionary should be used. For example, Thanksgiving can be defined as follows::

            recurrence = {
                'recurForEver': True,
                'recurYearlyByDay': {
                    'day': 'THURSDAY',
                    'week': 'FOURTH',
                    'month': 'NOVEMBER'
                }
            }

        The format of this dict, and accepted values can be found
        `here <https://developer.webex.com/docs/api/v1/
        webex-calling-organization-settings-with-location-scheduling/create-a-schedule-event>`_

        Args:
            name (str): The name of the holiday
            date (str): The date the holiday occurs this year in the format YYYY-MM-DD.
            recur (bool, optional): Whether the holiday recurs.
            recurrence (dict, optional): A dict of the recurrence info.

        Returns:
            bool: True on success, False otherwise

        Raises:
            TypeError: Raised when attempting to add a holiday to a businessHours schedule

        """
        log.debug(f"add_holiday() started")
        new_event = {"name": name,
                     "startDate": date, "endDate": date, "allDayEnabled": True}
        date_object = datetime.strptime(date, "%Y-%m-%d")
        # If this isn't a Holidays schedule, we shouldn't accept a new holiday
        if self.type.lower() != "holidays":
            log.debug("Trying to add a holiday to a non-holidays schedule. Raising exception.")
            raise TypeError("Schedule is not of type 'holidays'. Cannot add holiday.")
        if recur is True:
            if recurrence is not None:
                new_event['recurrence'] = recurrence
            else:
                new_event['recurrence'] = {"recurForEver": True,
                                           "recurYearlyByDate": {"dayOfMonth": date_object.strftime("%d"),
                                                                 "month": date_object.strftime("%B").upper()
                                                                 }
                                           }
        api_resp = webex_api_call("post", f"v1/telephony/config/locations/"
                                          f"{self.parent.id}/schedules/{self.type}/{self.id}/events", payload=new_event)
        if api_resp:
            self.refresh_config()
            return True
        else:
            return False

    def delete_event(self, event: str):
        """ Delete an event from within the LocationSchedule

        The event can be passed by name or by ID since both are unique.

        Args:
            event (str): The event name or event ID

        Returns:
            bool: True on success, False otherwise

        """
        log.debug("delete_event() started")
        # Since we accept both the event name or ID, we need to loop through the events to find the one we need.
        for e in self.config['events']:
            if e['name'] == event or e['id'] == event:
                api_resp = webex_api_call("delete", f"v1/telephony/config/locations/{self.parent.id}"
                                                    f"/schedules/{self.type}/{self.id}/events/{e['id']}")
                if api_resp is True:
                    # Get a new copy of the config
                    self.refresh_config()
                    return True
                else:
                    self.refresh_config()
                    return False
        return False

    def create_event(self, name: str, start_date: str, end_date: str, start_time: str = "",
                     end_time: str = "", all_day: bool = False, recurrence: Optional[dict] = None):
        """ Create a new event within the LocationSchedule instance

        Due to the flexibility of the event recurrence needs to be provided as a dict. The format of this dict, and
        accepted values can be found `here <https://developer.webex.com/docs/api/v1/webex-calling-organization-settings-with-location-scheduling/create-a-schedule-event>`_

        For example, the following dict would define an event that recurs every weekday, from 9:00-5:00::

            recurrence = {
                'recurForEver': True,
                'recurWeekly': {
                    'monday': True
                }
            }


        Args:
            name (str): The name of the event
            start_date (str): The start date of the event
            end_date (str): The end date of the event
            start_time (str, optional): If this is not an all-day event, the time the event starts
            end_time (str, optional): If this is not an all-day event, the time the event ends
            all_day (bool, optional): True if this is an all-day event. start_time and end_time will be ignored.
            recurrence (dict, optional): Dict of recurrence configuration

        Returns:
            bool: True on success, False otherwise

        """
        log.debug("create_event() started")
        # Input validation
        if all_day is False and (start_time == "" or end_time == ""):
            raise ValueError("If all_day == False, both start_time and end_time are required.")

        # Build the payload
        payload = {'name': name, 'startDate': start_date, 'endDate': end_date, 'allDay': all_day}
        if all_day is False:
            payload['startTime'] = start_time
            payload['endTime'] = end_time
        if recurrence is not None:
            payload['recurrence'] = recurrence

        api_resp = webex_api_call("post", f"v1/telephony/config/locations/{self.parent.id}/"
                                          f"schedules/{self.type}/{self.id}/events",
                                  payload=payload)
        if api_resp:
            self.refresh_config()
            return True
        else:
            self.refresh_config()
            return False

    def update_event(self, id: str, name: str = None, start_date: str = None, end_date: str = None,
                     start_time: str = None,
                     end_time: str = None, all_day: bool = None, recurrence: dict = None):
        log.debug("update_event() started")
        # Input validation
        if all_day is False and (start_time == "" or end_time == ""):
            raise ValueError("If all_day == False, both start_time and end_time are required.")

        # Build the payload
        payload = self.get_event_config_by_id(id)
        if name is not None:
            payload['name'] = name
        if start_date is not None:
            payload['startDate'] = start_date
        if end_date is not None:
            payload['endDate'] = end_date
        if all_day is not None:
            payload['allDay'] = all_day
            if all_day is False:
                if start_time is not None:
                    payload['startTime'] = start_time
                if end_time is not None:
                    payload['endTime'] = end_time
        if recurrence is not None:
            payload['recurrence'] = recurrence

        api_resp = webex_api_call("put", f"v1/telephony/config/")
        if api_resp:
            return True
        else:
            return False

    def get_event_config_by_id(self, id: str) -> Optional[dict]:
        """ Get the 'events' dict for a specific event.

        This method is useful if you are modifying an event and want to provide the full config.

        Args:
            id (str): The Event ID

        Returns:
            dict: The config of the event as a dict. If the Event ID is not found, None is returned.

        """
        for e in self.config['events']:
            if e['id'] == id:
                return e
        return None

    def clone(self, target_location: Optional[wxcadm.Location] = None, name: Optional[str] = None):
        """ Clone the LocationSchedule to another Location or to the same Location with a new name

        Either a `target_location` or a `name` must be specified. If both are specified, the Schedule will be cloned to
        the new Location with the given name.

        Args:
            target_location (Location, optional): The Location to clone the Schedule to
            name (str, optional): The name of the Schedule to create.

        Returns:
            str: The LocationSchedule ID of the newly-created Schedule

        Raises:
            ValueError: Raised when `target_location` and `name` are not provided

        """
        if target_location is None and name is None:
            raise ValueError('target_location or name must be provided')
        payload = dict(self.config)
        del payload['id']
        if name is not None:
            payload['name'] = name
        if target_location is not None:
            target_locid = target_location.id
        else:
            target_locid = self.parent.id
        response = webex_api_call('post', f'v1/telephony/config/locations/{target_locid}/schedules', payload=payload)
        return response['id']


class CallParkGroup:
    pass


@dataclass
class CallParkExtension:
    parent: wxcadm.location.Location
    """ The Location to which the Call Park Extension is assigned """
    id: str
    """ The ID of the Call Park Extension """
    name: str
    """ The name of the Call Park Extension """
    extension: str
    """ The Call Park Extension number (as a string) """

    def get_monitored_by(self):
        """ Returns a list of Users (Person) and Workspaces that are Monitoring this Park Extension """
        if isinstance(self.parent, wxcadm.Org):
            monitor_list = self.parent.get_all_monitoring()
        elif isinstance(self.parent, wxcadm.Location):
            monitor_list = self.parent.parent.get_all_monitoring()
        try:
            return monitor_list['park_extensions'][self.id]
        except KeyError:
            return None


@dataclass
class VoicePortal(RealtimeClass):
    """ The class for the Voice Portal at the Location

    .. note::

        Unlike all other methods, changing an attribute of this class will immediately push the change to Webex. This
        is being done to test a new API logic which performs real-time changes.

    """
    log.info("Getting Voice Portal data")
    location: wxcadm.Location = field(repr=False)
    """ The :class:`Location` of the Voice Portal """
    id: str = field(repr=True, init=False, default=None)
    """ The identifier for the Voice Portal """
    language: str = field(repr=True, init=False, default=None)
    """ The language to use for announcements """
    language_code: str = field(repr=True, init=False, default=None)
    """ Language code for voicemail group audio announcement """
    extension: str = field(repr=True, init=False, default=None)
    """ The Voice Portal extension """
    phone_number: str = field(repr=True, init=False, default=None)
    """ The Voice Portal DID """
    first_name: str = field(repr=True, init=False, default=None)
    """ The first name of the Voice Portal in the directory """
    last_name: str = field(repr=True, init=False, default=None)
    """ The last name of the Voice Portal in the directory """
    _api_fields = ['phone_number', 'extension']
    _initialized = False

    def __post_init__(self):
        self.data_url: str = f'v1/telephony/config/locations/{self.location.id}/voicePortal'
        super().__init__()

    def copy_config(self, target_location: wxcadm.Location, phone_number: str = None, passcode: str = None) -> bool:
        """ Copies the Voice Portal settings to another Location.

        The phone number and passcode will not be copied. To define those values as part of the copy operation,
        ensure those arguments are passed to the method.

        Args:
            target_location (wxcadm.Location): The :class:`Location` to copy the settings to
            phone_number (str, optional): The phone number to assign to the Voice Portal
            passcode (str, optional): The Voice Portal passcode to set

        Returns:
            bool: True on success

        """
        payload = {
            'name': 'Voice Portal',
            'languageCode': self.language_code,
            'extension': self.extension,
            'firstName': self.first_name,
            'lastName': self.last_name,
        }
        if phone_number is not None:
            payload['phoneNumber'] = phone_number
        if passcode is not None:
            payload['passcode'] = {
                'newPasscode': passcode,
                'confirmPasscode': passcode
            }
        if isinstance(target_location, wxcadm.Location):
            if target_location.calling_enabled is False:
                raise ValueError("Target Location is not Webex Calling enabled")
            webex_api_call('put', f"v1/telephony/config/locations/{target_location.id}/voicePortal",
                           payload=payload, params={'orgId': target_location.org_id})
        else:
            raise ValueError("target_location much be a Location instance")

        return True


class OutgoingPermissionDigitPatternList:
    def __init__(self, location: wxcadm.Location):
        log.info('Initializing OutgoingPermissionDigitPatternList')
        self.location = location
        self.patterns = self._get_data()

    def _get_data(self) -> list:
        pattern_list = []
        log.debug('Getting data from Webex')
        response = webex_api_call('get',
                                  f'v1/telephony/config/locations/{self.location.id}/outgoingPermission/digitPatterns')
        for pattern in response['digitPatterns']:
            pattern_list.append(OutgoingPermissionDigitPattern(self.location, config=pattern))
        return pattern_list

    def refresh(self):
        self.patterns = self._get_data()
        return self

    def create(self, name: str, pattern: str, action: str, transfer_enabled: Optional[bool] = False):
        """ Create a new Outgoing Permission Digit Pattern for the Location

        Args:
            name (str): The name of the pattern

            pattern (str): The digit pattern. See Webex documentation for valid values.

            action (str): The action to take. Valid values are `'ALLOW'`, `'BLOCK'`, `'AUTH_CODE'`,
            `'TRANSFER_NUMBER_1'`, `'TRANSFER_NUMBER_2'`, `'TRANSFER_NUMBER_3'`

            transfer_enabled (bool): Whether the setting is used for transferred or forwarded calls.

        Returns:
            OutgoingPermissionDigitPattern: The created pattern

        Raises:
            wxcadm.APIError: Raised when the API call fails

        """
        payload = {
            'name': name,
            'pattern': pattern,
            'action': action,
            'transferEnabled': transfer_enabled
        }
        response = webex_api_call('post',
                                  f'v1/telephony/config/locations/{self.location.id}/outgoingPermission/digitPatterns',
                                  params={'orgId': self.location.org_id},
                                  payload=payload)
        # Use the ID that was returned to get the object we want to return
        pattern_id = response['id']
        response = webex_api_call(
            'get',
            f'v1/telephony/config/locations/{self.location.id}/outgoingPermission/digitPatterns/{pattern_id}',
            params={'orgId': self.location.org_id}
        )
        new_pattern = OutgoingPermissionDigitPattern(self.location, response)
        self.patterns.append(new_pattern)
        return new_pattern

    def delete_all(self):
        """ Delete all Outgoing Permission Digit Patterns for this Location

        Returns:
            bool: True on success

        """
        webex_api_call(
            'delete',
            f'v1/telephony/config/locations/{self.location.id}/outgoingPermission/digitPatterns',
            params={'orgId': self.location.org_id}
        )
        self.refresh()
        return True


class OutgoingPermissionDigitPattern:
    def __init__(self, location: wxcadm.Location, config: dict):
        self.location = location
        self.id: str = config.get('id', '')
        self.name: str = config.get('name', '')
        self.pattern: str = config.get('pattern', '')
        self.action: str = config.get('action', '')
        self.transfer_enabled: bool = config.get('transferEnabled')

    def delete(self) -> bool:
        """ Delete the specified digit pattern

        Returns:
            bool: True on success

        """
        webex_api_call(
            'delete',
            f'v1/telephony/config/locations/{self.location.id}/outgoingPermission/digitPatterns/{self.id}'
        )
        return True

    def update(self,
               name: Optional[str] = None,
               pattern: Optional[str] = None,
               action: Optional[str] = None,
               transfer_enabled: Optional[bool] = None):
        name = self.name if name is None else name
        pattern = self.pattern if pattern is None else pattern
        action = self.action if action is None else action
        transfer_enabled = self.transfer_enabled if transfer_enabled is None else transfer_enabled
        payload = {
            'name': name,
            'pattern': pattern,
            'action': action,
            'transferEnabled': transfer_enabled
        }
        webex_api_call(
            'put',
            f'v1/telephony/config/locations/{self.location.id}/outgoingPermission/digitPatterns/{self.id}',
            params={'orgId': self.location.org_id},
            payload=payload
        )
        self.name = name
        self.pattern = pattern
        self.action = action
        self.transfer_enabled = transfer_enabled
        return self


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class VoicemailGroup:
    org: wxcadm.Org
    id: str
    name: str
    location_name: str
    location_id: str
    extension: str
    enabled: bool
    esn: Optional[str] = None
    _phone_number: Optional[str] = field(init=False, default=None)
    _first_name: Optional[str] = field(init=False, default=None)
    _last_name: Optional[str] = field(init=False, default=None)
    _language_code: Optional[str] = field(init=False, default=None)
    _greeting: Optional[str] = field(init=False, default=None)
    _greeting_uploaded: Optional[bool] = field(init=False, default=None)
    _greeting_description: Optional[str] = field(init=False, default=None)
    _message_storage: Optional[dict] = field(init=False, default=None)
    _notifications: Optional[dict] = field(init=False, default=None)
    _fax_message: Optional[dict] = field(init=False, default=None)
    _transfer_to_number: Optional[dict] = field(init=False, default=None)
    _email_copy_of_message: Optional[dict] = field(init=False, default=None)
    _message_forwarding_enabled: Optional[bool] = field(init=False, default=None)

    def _get_details(self):
        response = webex_api_call('get',
                                  f'v1/telephony/config/locations/{self.location_id}/voicemailGroups/{self.id}',
                                  params={'orgId': self.org.id})
        self.name = response.get('name')
        self.extension = response.get('extension')
        self.enabled = response.get('enabled')
        self._phone_number = response.get('phoneNumber', None)
        self._first_name = response.get('firstName', '')
        self._last_name = response.get('lastName', '')
        self._language_code = response.get('languageCode', '')
        self._greeting = response.get('greeting', '')
        self._greeting_uploaded = response.get('greetingUploaded', '')
        self._greeting_description = response.get('greetingDescription', '')
        self._message_storage = response.get('messageStorage', None)
        self._notifications = response.get('notifications', None)
        self._fax_message = response.get('faxMessage', None)
        self._transfer_to_number = response.get('transferToNumber', None)
        self._email_copy_of_message = response.get('emailCopyOfMessage', None)
        self._message_forwarding_enabled = response.get('voiceMessageForwardingEnabled', None)

    @property
    def phone_number(self):
        """ The phone number of the Voicemail Group """
        if self._phone_number is None:
            self._get_details()
        return self._phone_number

    @property
    def last_name(self):
        """ The last name of the Voicemail Group """
        if self._last_name is None:
            self._get_details()
        return self._last_name

    @property
    def first_name(self):
        """ The firt name of the Voicemail Group """
        if self._first_name is None:
            self._get_details()
        return self._first_name

    @property
    def language_code(self):
        """ The language code of the Voicemail Group """
        if self._language_code is None:
            self._get_details()
        return self._language_code

    @property
    def greeting_type(self):
        """ The greeting type, either 'DEFAULT' or 'CUSTOM' """
        if self._greeting is None:
            self._get_details()
        return self._greeting

    @property
    def greeting_uploaded(self):
        """ True if CUSTOM greeting has been previously uploaded """
        if self._greeting_uploaded is None:
            self._get_details()
        return self._

    @property
    def greeting_description(self):
        """ CUSTOM greeting for previously uploaded greeting """
        if self._greeting_description is None:
            self._get_details()
        return self._greeting_description

    @property
    def message_storage(self):
        """ Message storage information as a dict """
        if self._message_storage is None:
            self._get_details()
        return self._message_storage

    @property
    def notifications(self):
        """ Notification settings as a dict """
        if self._notifications is None:
            self._get_details()
        return self._notifications

    @property
    def fax_message(self):
        """ Fax messaging settings as a dict """
        if self._fax_message is None:
            self._get_details()
        return self._fax_message

    @property
    def transfer_to_number(self):
        """ Transfer settings for the Voicemail Group as a dict """
        if self._transfer_to_number is None:
            self._get_details()
        return self._transfer_to_number

    @property
    def email_copy_of_message(self):
        """ Email CC settings as a dict """
        if self._email_copy_of_message is None:
            self._get_details()
        return self._email_copy_of_message

    @property
    def message_forwarding_enabled(self):
        """ Whether voice message forwarding is enabled """
        if self._message_forwarding_enabled is None:
            self._get_details()
        return self._message_forwarding_enabled

    def update(self,
               name: Optional[str] = None,
               phone_number: Optional[str] = None,
               extension: Optional[str] = None,
               first_name: Optional[str] = None,
               last_name: Optional[str] = None,
               enabled: Optional[bool] = None,
               language_code: Optional[str] = None,
               greeting_type: Optional[str] = None,
               greeting_description: Optional[str] = None,
               message_storage: Optional[dict] = None,
               notifications: Optional[dict] = None,
               fax_message: Optional[dict] = None,
               transfer_to_number: Optional[dict] = None,
               email_copy_of_message: Optional[dict] = None
               ):
        """ Update Voicmeail Group settings

        When using this method, pass only the values that you wish to change. Values omitted from the parameters will
        not be changed.

        Args:
            name (str, optional): The name of the Voicemail Group
            phone_number (str, optional): The phone number of the Voicemail Group
            extension (str, optional): The extension of the Voicemail Group
            first_name (str, optional): The first name of the Voicemail Group
            last_name (str, optional): The last name of the Voicemail Group
            enabled (bool, optional): Whether the Voicemail Group is enabled
            language_code (str, optional): The language code
            greeting_type (str, optional): The greeting type, either 'CUSTOM' or 'DEFAULT'
            greeting_description (str, optional): The greeting for CUSTOM greetings
            message_storage (dict, optional): The message storage settings as a dict
            notifications (dict, optional): The notification settings as a dict
            fax_message (dict, optional): The fax settings as a dict
            transfer_to_number (dict, optional): The transfer settings as a dict
            email_copy_of_message (dict, optional): The email CC settings as a dict

        Returns:
            VoicemailGroup: The updated VoicemailGroup instance

        Raises:
            wxcadm.APIError: Raised when the update is rejected by Webex

        """
        name = self.name if name is None else name
        phone_number = self.phone_number if phone_number is None else phone_number
        extension = self.extension if extension is None else extension
        first_name = self.first_name if first_name is None else first_name
        last_name = self.last_name if last_name is None else last_name
        enabled = self.enabled if enabled is None else enabled
        language_code = self.language_code if language_code is None else language_code
        greeting_type = self.greeting_type if greeting_type is None else greeting_type
        greeting_description = self.greeting_description if greeting_description is None else greeting_description
        message_storage = self.message_storage if message_storage is None else message_storage
        notifications = self.notifications if notifications is None else notifications
        fax_message = self.fax_message if fax_message is None else fax_message
        transfer_to_number = self.transfer_to_number if transfer_to_number is None else transfer_to_number
        email_copy_of_message = self.email_copy_of_message if email_copy_of_message is None else email_copy_of_message
        payload = {
            'name': name,
            'phoneNumber': phone_number,
            'extension': extension,
            'firstName': first_name,
            'lastName': last_name,
            'enabled': enabled,
            'languageCode': language_code,
            'greeting': greeting_type,
            'greetingDescription': greeting_description,
            'messageStorage': message_storage,
            'notifications': notifications,
            'faxMessage': fax_message,
            'transferToNumber': transfer_to_number,
            'emailCopyOfMessage': email_copy_of_message
        }
        webex_api_call('put',
                       f'v1/telephony/config/locations/{self.location_id}/voicemailGroups/{self.id}',
                       payload=payload)
        self._get_details()
        return self

    def delete(self):
        """ Delete the Group Voicemail """
        webex_api_call('delete',
                       f'v1/telephony/config/locations/{self.location_id}/voicemailGroups/{self.id}')
        return True

    def enable_email_copy(self, email: str):
        """ Enable sending of copies of voicemails to an email

        This is a shortcut for calling the :meth:`update()` method with the `email_copy_of_message` argument dict

        Args:
             email (str): The email address to send copies of messages to

        Returns:
            VoicemailGroup: The updated VoicemailGroup instance

        Raises:
            wxcadm.APIError: Raised when the update is rejected by Webex

        """
        email_conf = {'enabled': True, 'emailId': email}
        self.update(email_copy_of_message=email_conf)
        return self


class VoicemailGroupList(UserList):
    def __init__(self, org: wxcadm.Org):
        log.info(f'Initializing VoicemailGroupList for Org: {org.name}')
        super().__init__()
        self.org = org
        self.data = self._get_data()

    def _get_data(self) -> list:
        log.debug('Getting data')
        data = []
        response = webex_api_call('get', 'v1/telephony/config/voicemailGroups', params={'orgId': self.org.id})
        for group in response['voicemailGroups']:
            group['org'] = self.org
            data.append(VoicemailGroup.from_dict(group))
        return data

    def refresh(self):
        self.data = self._get_data()
        return self

    def get(self, name: Optional[str] = None, id: Optional[str] = None):
        """ Get the VoicemailGroup matching the given `id` or `name`

        Args:
             name (str, optional): The Voicemail Group name to retrieve
             id (str, optional): The Voicemail Group ID to retrieve

        Returns:
            VoicemailGroup: The matching instance. None is returned if no match is found

        """
        for item in self.data:
            if item.id == id or item.name == name:
                return item
        return None

    def create(self,
               location: wxcadm.Location,
               name: str,
               extension: str,
               passcode: str,
               phone_number: Optional[str] = None,
               first_name: Optional[str] = None,
               last_name: Optional[str] = None,
               language_code: Optional[str] = 'en_us',
               message_storage: Optional[dict] = None,
               notifications: Optional[dict] = None,
               fax_message: Optional[dict] = None,
               transfer_to_number: Optional[dict] = None,
               email_copy_of_message: Optional[dict] = None):
        log.info(f'Creating Voicemail Group {name} in Location {location.name}')
        if message_storage is None:
            message_storage = {'storageType': 'INTERNAL'}
        if notifications is None:
            notifications = {'enabled': False}
        if fax_message is None:
            fax_message = {'enabled': False}
        if transfer_to_number is None:
            transfer_to_number = {'enabled': False}
        if email_copy_of_message is None:
            email_copy_of_message = {'enabled': False}

        payload = {
            'name': name,
            'extension': extension,
            'passcode': passcode,
            'phoneNumber': phone_number,
            'firstName': first_name,
            'lastName': last_name,
            'languageCode': language_code,
            'messageStorage': message_storage,
            'notifications': notifications,
            'faxMessage': fax_message,
            'transferToNumber': transfer_to_number,
            'emailCopyOfMessage': email_copy_of_message
        }
        response = webex_api_call('post',
                                  f'v1/telephony/config/locations/{location.id}/voicemailGroups',
                                  params={'orgId': self.org.id},
                                  payload=payload)
        vg_id = response['id']
        self.refresh()
        return self.get(id=vg_id)




from __future__ import annotations

import requests
from collections import UserList
from typing import Optional, Union, TYPE_CHECKING
if TYPE_CHECKING:
    from wxcadm import Org
import wxcadm.location
import wxcadm
from wxcadm import log
from .common import *
from .exceptions import *
from .device import DeviceList
from .monitoring import MonitoringList


class WorkspaceLocationList(UserList):
    def __init__(self, parent: wxcadm.Org):
        """

        .. deprecated:: 3.4.0
            The Workspace Location concept is being removed from the Webex APIs. All Workspace Locations are now simply
            :class:`Location`s now. This class will still show up in come cases until it is completely removed.

        """
        super().__init__()
        log.debug("Initializing WorkspaceLocationList")
        self.parent: wxcadm.Org = parent
        self.data: list = self._get_workspace_locations()

    def _get_workspace_locations(self):
        """

        .. deprecated:: 3.4.0
            The Workspace Location concept is being removed from the Webex APIs. All Workspace Locations are now simply
            :class:`Location`s now. This class will still show up in come cases until it is completely removed.

        """
        log.debug("Getting list of Workspace Locations")
        workspace_locations = []
        log.debug(f"Using Org {self.parent.name} as WorkspaceLocation filter")
        params = {'orgId': self.parent.id}

        response = webex_api_call("get", "v1/workspaceLocations", params=params)
        log.debug(f"Received {len(response)} Workspace Locations from Webex")
        for entry in response:
            workspace_locations.append(WorkspaceLocation(self.parent, id=entry['id'], config=entry))
        return workspace_locations

    def get_by_id(self, id: str):
        """ Get a WorkspaceLocation instance from the WorkspaceLocationList by ID

        Args:
            id (str): The WorkspaceLocation ID to find

        Returns:
            WorkspaceLocation: The :py:class:`WorkspaceLocation` instance for the given ID.
            None is returned if no match is found.

        """
        entry: WorkspaceLocation
        for entry in self.data:
            if entry.id == id:
                return entry
        return None

    def get(self, id: Optional[str] = None, name: Optional[str] = None):
        """ Get a WorkspaceLocation instance by ID or name

        Args:
            id (str, optional): The WorkspaceLocation ID
            name (str, optional): The WorkspaceLocation name

        Returns:
            WorkspaceLocation: The :class:`WorkspaceLocation` for the given value.
            None is returned if no match is found.

        """
        if name is None and id is None:
            raise ValueError("Method requires id or name argument")
        entry: WorkspaceLocation
        for entry in self.data:
            if entry.id == id or entry.name == name:
                return entry
        return None

class WorkspaceList(UserList):
    def __init__(self, parent: Union[wxcadm.Org, wxcadm.WorkspaceLocation, wxcadm.Location]):
        """

        .. deprecated:: 3.4.0
            The Workspace Location concept is being removed from the Webex APIs. All Workspace Locations are now simply
            :class:`Location`s now. This class will still show up in come cases until it is completely removed.

        """

        super().__init__()
        log.debug("Initializing WorkspaceList instance")
        self.parent: Union[wxcadm.Org, wxcadm.Location] = parent
        self.data: list = self._get_workspaces()

    def _get_workspaces(self):
        log.debug("Getting List of Workspaces")
        workspaces = []
        if isinstance(self.parent, wxcadm.location.Location):
            log.debug(f"Using Location {self.parent.name} as Workspace filter")
            params = {'locationId': self.parent.id}
        elif isinstance(self.parent, wxcadm.workspace.WorkspaceLocation):
            log.debug(f"Using WorkspaceLocation {self.parent.name} as Workspace filter")
            params = {'workspaceLocationId': self.parent.id}
        elif isinstance(self.parent, wxcadm.org.Org):
            log.debug(f"Using Org {self.parent.name} as Workspace filter")
            params = {'orgId': self.parent.id}
        else:
            params = {}
        response = webex_api_call("get", "v1/workspaces", params=params)
        log.debug(f"Received {len(response)} Workspaces from Webex")
        for entry in response:
            workspaces.append(Workspace(self.parent, id=entry['id'], config=entry))
        return workspaces

    def refresh(self):
        """ Re-query the list of Workspaces from Webex """
        self.data: list = self._get_workspaces()

    def get_by_id(self, id: str):
        """ Get a Workspace instance from the WorkspaceList by ID

        Args:
            id (str): The Workspace ID to find

        Returns:
            Workspace: The :py:class:`Workspace` instance for the given ID. None is returned if no match is found.

        """
        entry: Workspace
        for entry in self.data:
            if entry.id == id:
                return entry
        return None

    def get(self, id: Optional[str] = None, name: Optional[str] = None, uuid: Optional[str] = None):
        """ Get a Workspace instance by ID, Name or UUID

        Args:
            id (str, optional): The Workspace ID
            name (str, optional): The Workspace name
            uuid (str, optional): The Workspace UUID

        Returns:
            Workspace: The :class:`Workspace` for the given criteria

        """
        entry: Workspace
        for entry in self.data:
            if entry.id == id or (entry.name is not None and entry.name == name):
                return entry
            if uuid is not None:
                if entry.spark_id.split('/')[-1].upper() == uuid.upper():
                    return entry
        return None

    def webex_calling(self) -> list:
        """ Return a list of Workspaces that have Webex Calling enabled """
        wxc_workspaces = []
        for entry in self.data:
            if entry.calling.lower() == 'webexcalling':
                wxc_workspaces.append(entry)
        return wxc_workspaces

    def create(self, location: wxcadm.Location,
               name: str,
               floor: Optional[WorkspaceLocationFloor] = None,
               capacity: Optional[int] = None,
               type: Optional[str] = 'notSet',
               phone_number: Optional[str] = None,
               extension: Optional[str] = None,
               notes: Optional[str] = None,
               hotdesking: Optional[bool] = False,
               supported_devices: Optional[str] = 'phones',
               license_type: str = 'workspace'
               ):
        """ Create a new Workspace

        In order to enable Webex Calling, a Location must be provided as well as an extension, phone number, or both.

        Args:
            location (Location): The :class:`Location` where the Workspace will be located
            name (str): The name of the Workspace
            floor (LocationFloor, optional): The :class:`LocationFloor` where the Workspace is located
            capacity (int, optional): The capacity of the Workspace
            type (str, optional): The type of Workspace. See the developer documentation for options. Defaults to `noteSet`
            phone_number (str, optional): The Webex Calling phone number for the Workspace
            extension (str, optional): The Webex Calling extension for the Workspace
            notes (str, optional): Free-form text notes
            hotdesking (bool, optional): Whether the workspace is enabled for Hot Desking. Defaults to False.
            supported_devices (str, optional): `phones` or `collaborationDevices`. Defaults to `phones`
            license_type (str, optional): 'workspace' or 'professional'. Defaults to 'workspace'

        Returns:
            Workspace: The :class:`Workspace` instance that is created in Control Hub

        Raises:
            ValueError: Raised when an ``extension`` or ``phone_number`` is not provided.

        """
        if extension is None and phone_number is None:
            raise ValueError("Must provide extension, phone_number, or both")
        # Removed 3.4.0 when Workspace Locations were deprecated
        # if location.workspace_location is None:
        #     raise KeyError(f"Location {location.name} does not have a Workspace Location")
        if hotdesking is True:
            hotdesking = 'on'
        else:
            hotdesking = 'off'

        payload = {
            'orgId': self.parent.org_id,
            'displayName': name,
            'locationId': location.id,
            'floorId': floor,
            'capacity': capacity,
            'type': type,
            'calling': {
                'type': 'webexCalling',
                'webexCalling': {
                    'phoneNumber': phone_number,
                    'extension': extension,
                    'locationId': location.id
                }
            },
            'notes': notes,
            'hotdeskingStatus': hotdesking,
            'supportedDevices': supported_devices
        }

        # Find an available license if a Professional license was requested
        if license_type.lower() == 'professional':
            wxc_license = self.parent.get_wxc_person_license()
            payload['calling']['webexCalling']['licenses'] = [wxc_license]

        response = webex_api_call('post', 'v1/workspaces', payload=payload)
        new_workspace_id = response['id']
        # TODO - Eventually I would like to remove the refresh() and just add the new Workspace directly
        self.parent.workspaces.refresh()
        new_workspace = self.parent.workspaces.get_by_id(new_workspace_id)
        return new_workspace


class Workspace:
    def __init__(self, parent: wxcadm.Org, id: str, config: Optional[dict] = None):
        """Initialize a Workspace instance

        If only the `id` is provided, the configuration will be fetched from
        the Webex API. To save API calls, the config dict can be passed using the `config` argument

        Args:
            parent (Org): The Organization to which this workspace belongs
            id (str): The Webex ID of the Workspace
            config (dict): The configuration of the Workspace as returned by the Webex API

        """
        self.id: str = id
        """The Webex ID of the Workspace"""
        self._parent: Union[wxcadm.Org, wxcadm.Location] = parent
        # Instance attributes
        self.location: Optional[wxcadm.Location] = None
        """ The :class:`Location` of the Workspace """
        self.floor_id: Optional[str] = None
        """The Webex ID of the Location Floor ID"""
        self.name: str = ""
        """The name of the Workspace"""
        self.capacity: Optional[int] = None
        """The capacity of the Workspace"""
        self.type: Optional[str] = None
        """
        The type of Workspace. Valid values are:

            "notSet": No value set
            "focus": High concentration
            "huddle": Brainstorm/collaboration
            "meetingRoom": Dedicated meeting space
            "open": Unstructured agile
            "desk": Individual
            "other": Unspecified
        """
        self.sip_address: Optional[str] = None
        """The SIP Address used to call to the Workspace"""
        self.created: Optional[str] = None
        """The date and time the workspace was created"""
        self.calling: Optional[str] = None
        """
        The type of Calling license assigned to the Workspace. Valid values are:

            'freeCalling': Free Calling
            'hybridCalling': Hybrid Calling
            'webexCalling': Webex Calling
            'webexEdgeForDevices': Webex Edge for Devices
        """
        self.calendar: Optional[dict] = None
        """The type of calendar connector assigned to the Workspace"""
        self.notes: Optional[str] = None
        """Notes associated with the Workspace"""
        self.licenses: list = []
        """ The licenses for the Workspace when the calling value is 'webexCalling' """

        # Property storage
        self._numbers: Optional[dict] = None
        self._devices: Optional[DeviceList] = None
        self._monitoring: Optional[dict] = None


        if config:
            self.__process_config(config)
        else:
            self.get_config()

    def __str__(self):
        return self.name

    @property
    def org_id(self) -> str:
        """ The Org ID of the Workspace """
        return self._parent.org_id

    @property
    def number(self):
        """ The phone number of the Workspace when the Calling type is 'webexCalling'

        .. note::

            When only one number is assigned, a string will be returned. If multiple numbers are assigned, a list
            of strings will be returned.

        """
        if self._numbers is None:
            log.info(f"Getting numbers for Workspace: {self.name}")
            self._numbers = webex_api_call('get', f"v1/workspaces/{self.id}/features/numbers",
                                           params={'orgId': self.org_id})
            log.debug(self._numbers)
        if len(self._numbers['phoneNumbers']) == 1:
            return self._numbers['phoneNumbers'][0].get('external', None)
        elif len(self._numbers['phoneNumbers']) > 1:
            num_list = []
            for number in self._numbers['phoneNumbers']:
                if number['external'] is not None:
                    num_list.append(number['external'])
            return num_list
        else:
            return None

    @property
    def extension(self):
        """ The extension of the Workspace when the Calling type is 'webexCalling' """
        if self._numbers is None:
            self._numbers = webex_api_call('get', f"v1/workspaces/{self.id}/features/numbers",
                                           params={'orgId': self.org_id})
        if len(self._numbers['phoneNumbers']) == 1:
            return self._numbers['phoneNumbers'][0].get('extension', None)
        elif len(self._numbers['phoneNumbers']) > 1:
            num_list = []
            for number in self._numbers['phoneNumbers']:
                if number['extension'] is not None:
                    num_list.append(number['extension'])
            return num_list
        else:
            return None

    @property
    def esn(self):
        """ The Enterprise Significant Number (ESN) for the Workspace when the Calling type is 'webexCalling' """
        if self._numbers is None:
            self._numbers = webex_api_call('get', f"v1/workspaces/{self.id}/features/numbers",
                                           params={'orgId': self.org_id})
        if len(self._numbers['phoneNumbers']) == 1:
            return self._numbers['phoneNumbers'][0].get('esn', None)
        elif len(self._numbers['phoneNumbers']) > 1:
            num_list = []
            for number in self._numbers['phoneNumbers']:
                if number['esn'] is not None:
                    num_list.append(number['esn'])
            return num_list
        else:
            return None

    @property
    def spark_id(self):
        """ The internal identifier used by Webex """
        return decode_spark_id(self.id)

    @property
    def monitoring(self) -> MonitoringList:
        """ The :class:`~.monitoring.MonitoringList` associated with the Workspace """
        if self._monitoring is None:
            response = webex_api_call("get", f"v1/workspaces/{self.id}/features/monitoring",
                                      params={"orgId": self.org_id})
            response['parent'] = self
            response['org'] = self._parent
            if "monitoredElements" not in response.keys():
                response['monitoredElements'] = []
            self._monitoring = MonitoringList.from_dict(response)
        return self._monitoring

    def get_monitored_by(self):
        """ Returns a list of Users (Person) and Workspaces that are Monitoring this Workspace """
        if isinstance(self._parent, wxcadm.Org):
            monitor_list = self._parent.get_all_monitoring()
        elif isinstance(self._parent, wxcadm.Location):
            monitor_list = self._parent.parent.get_all_monitoring()
        try:
            return monitor_list['workspaces'][self.id]
        except (KeyError, TypeError):
            return None

    def delete(self):
        """ Delete the Workspace

        Returns:
            bool: True on success. In failure conditions, an exception will be raised

        Raises:
            APIError: Raised when the API call fails for any reason

        """
        response = webex_api_call('delete', f'v1/workspaces/{self.id}')
        self._parent.workspaces.refresh()
        return True

    @property
    def devices(self):
        if self._devices is None:
            self._devices = DeviceList(self)
        return self._devices

    @property
    def caller_id(self) -> dict:
        """ The Caller ID settings for the Workspace """
        log.info(f"Getting Caller ID config for Workspace: {self.name}")
        response = webex_api_call('get', f"v1/workspaces/{self.id}/features/callerId", params={'orgId': self.org_id})
        return response

    def set_caller_id(self, name: str, number: str, block_for_received_calls: Optional[bool] = False):
        """ Set the Caller ID for the Workspace

        Args:
            name (str): The name to set as the Caller ID Name. Also accepts keywords: ``direct`` sets the name to the
                Workspace's name . ``location`` sets the name to the name of the Location.
            number (str): The number to set as the Caller ID.  Also accepts keywords: ``direct`` sets the number to the
                Workspace's DID in Webex. ``location`` sets the name to the main number of the Location.
            block_for_received_calls (bool, optional): Block this Workspace's identity when receiving a call


        Returns:
            bool: True on success

        Raises:
            wxcadm.exceptions.APIError: Raised when there is a problem with the API call

        """
        log.info(f"Setting Caller ID for {self.name}")
        log.debug(f"\tNew Name: {name}\tNew Number: {number}")
        payload = {}
        # Handle the possible number values
        if number.lower() == "direct":
            payload['selected'] = "DIRECT_LINE"
        elif number.lower() == "location":
            payload['selected'] = "LOCATION_NUMBER"
        else:
            payload['selected'] = "CUSTOM"
            payload['customNumber'] = number
        # Then deal with possible name values
        if name.lower() == "direct":
            payload['externalCallerIdNamePolicy'] = "DIRECT_LINE"
        elif name.lower() == "location":
            payload['externalCallerIdNamePolicy'] = "LOCATION"
        else:
            payload['externalCallerIdNamePolicy'] = "OTHER"
            payload['customExternalCallerIdName'] = name
        payload['blockInForwardCallsEnabled'] = block_for_received_calls

        webex_api_call('put', f"v1/workspaces/{self.id}/features/callerId", params={'orgId': self.org_id},
                       payload=payload)
        return True

    def get_config(self):
        """Get (or refresh) the confirmation of the Workspace from the Webex API"""
        log.info(f"Getting Workspace config for {self.id}")
        r = requests.get(_url_base + f"v1/workspaces/{self.id}", headers=self._headers,
                         params={'orgId': self._parent.org_id})
        if r.status_code in [200]:
            response = r.json()
            self.__process_config(response)
        else:
            raise APIError(f"Unable to fetch workspace config for {self.id}")

    def __process_config(self, config: dict):
        """Processes the config dict, whether passed in init or from an API call"""
        self.name = config.get("displayName", "")
        if 'locationId' in config.keys():
            if isinstance(self._parent, wxcadm.Org):
                self.location = self._parent.locations.get(id=config['locationId'])
            elif isinstance(self._parent, wxcadm.Location):
                self.location = self._parent
        else:
            self.location = config.get("workspaceLocationId", None)
        self.floor = config.get("floorId", "")
        self.capacity = config.get("capacity", 0)
        if 'type' in config:
            self.type = config['type']
        self.sip_address = config.get("sipAddress", "")
        self.created = config.get("created", "")
        if "calling" in config:
            self.calling = config['calling']['type']
        else:
            self.calling = "None"
        self.calendar = config.get('calendar', None)
        self.notes = config.get("notes", "")
        if self.calling == 'webexCalling':
            self.licenses = config['calling']['webexCalling'].get('licenses', [])

    @property
    def ecbn(self) -> dict:
        """ The Emergency Callback Number details of the Workspace """
        response = webex_api_call('get', f'v1/telephony/config/workspaces/{self.id}/emergencyCallbackNumber',
                                  params={'orgId': self.org_id})
        return response

    def set_ecbn(self, value: Union[str, wxcadm.Person, wxcadm.Workspace, wxcadm.VirtualLine]):
        """ Set the ECBN of the Workspace

        Valid values are ``'direct'``, ``'location'``, or a :class:`Person`, :class:`Workspace`, or
        :class:`VirtualLine` to set the ECBN to one of those.

        Args:
            value (str, Person, Workspace, VirtualLine): The value to set the ECBN to

        Returns:
            bool: True on success

        """
        if isinstance(value, wxcadm.Person) or \
                isinstance(value, wxcadm.Workspace) or \
                isinstance(value, wxcadm.VirtualLine):
            payload = {
                'selected': 'LOCATION_MEMBER_NUMBER',
                'locationMemberId': value.id
            }
        elif value.lower() == 'direct' or value.lower() == 'direct_line':
            payload = {'selected': 'DIRECT_LINE'}
        elif value.lower() == 'location' or value.lower() == 'location_ecbn':
            payload = {'selected': 'LOCATION_ECBN'}
        else:
            raise ValueError('Unknown value')

        response = webex_api_call('put', f'v1/telephony/config/workspaces/{self.id}/emergencyCallbackNumber',
                                  params={'orgId': self.org_id}, payload=payload)
        return response

    @property
    def license_type(self) -> Optional[str]:
        """ The type of Webex Calling license used by the Workspace. Either 'WORKSPACE' or 'PROFESSIONAL'.

        Returns:
            str: The type of license, either 'WORKSPACE' or 'PROFESSIONAL'

        """
        log.info(f"Getting license type for Workspace: {self.name}")
        if len(self.licenses) == 1:
            lic_name = self._parent.get_license_name(self.licenses[0])
            if 'Workspaces' in lic_name:
                return 'WORKSPACE'
            elif 'Professional' in lic_name:
                return 'PROFESSIONAL'
            else:
                return lic_name
        if len(self.licenses) == 0:
            return None
        else:
            return "MULTIPLE"

    def set_professional_license(self):
        if self.license_type.upper() == 'PROFESSIONAL':
            return True
        new_license = self._parent.get_wxc_person_license()
        payload = {
            'displayName': self.name,
            'locationId': self.location.id,
            'floorId': self.floor_id,
            'capacity': self.capacity,
            'type': self.type,
            'calling': {
                'type': 'webexCalling',
                'webexCalling': {
                    'licenses': [
                        new_license
                    ]
                }
            },
            'notes': self.notes,


        }
        response = webex_api_call('put', f"v1/workspaces/{self.id}",
                                  params={'orgId': self.org_id},
                                  payload=payload)
        self.__process_config(response)
        return True


class WorkspaceLocation:
    def __init__(self, parent: wxcadm.Org, id: str, config: dict = None):
        """Initialize a WorkspaceLocation instance.

        If only the `id` is provided, the configuration will be fetched from
            the Webex API. To save API calls, the config dict can be passed using the `config` argument

        Args:
            parent (Org): The Organization to which this WorkspaceLocation belongs
            id (str): The Webex ID of the WorkspaceLocation
            config (dict): The configuration of the WorkspaceLocation as returned by the Webex API

        .. deprecated:: 3.4.0
            The Workspace Location concept is being removed from the Webex APIs. All Workspace Locations are now simply
            :class:`Location`s now. This class will still show up in come cases until it is completely removed.

        """
        self.id: str = id
        """The Webex ID of the Workspace"""
        self._parent: wxcadm.Org = parent
        # Attributes inherited from the Org parent
        self._headers = self._parent._headers
        self._params = self._parent._params
        # Instance attributes
        self.name: Optional[str] = None
        """The name of the WorkspaceLocation"""
        self.address: Optional[str] = None
        """The address of the WorkspaceLocation"""
        self.country: Optional[str] = None
        """The country code (ISO 3166-1) for the WorkspaceLocation"""
        self.city: Optional[str] = None
        """The city name where the WorkspaceLocation is located"""
        self.latitude: Optional[float] = None
        """The WorkspaceLocation latitude"""
        self.longitude: Optional[float] = None
        """The WorkspaceLocation longitude"""
        self.notes: Optional[str] = None
        """Notes associated with the WorkspaceLocation"""
        self.floors: Optional[list] = None
        self._workspaces: Optional[WorkspaceList] = None

        if config:
            self.__process_config(config)
        else:
            self.get_config()

    def get_config(self):
        """Get (or refresh) the configuration of the WorkspaceLocations from the Webex API"""
        log.info(f"Getting Workspace config for {self.id}")
        r = requests.get(_url_base + f"v1/workspaceLocations/{self.id}", headers=self._headers, params=self._params)
        if r.status_code in [200]:
            response = r.json()
            self.__process_config(response)
        else:
            raise APIError(f"Unable to fetch workspace config for {self.id}")

    def get_floors(self):
        """Get (or refresh) the WorkspaceLocationFloor instances for this WorkspaceLocation"""
        log.info(f"Getting Location Floors for {self.name}")
        self.floors = []
        r = requests.get(_url_base + f"v1/workspaceLocations/{self.id}/floors",
                         headers=self._headers, params=self._params)
        response = r.json()
        for floor in response['items']:
            this_floor = WorkspaceLocationFloor(floor)
            self.floors.append(this_floor)

    @property
    def workspaces(self):
        """ The :py:class:`Workspace` instances for this WorkspaceLocation """
        if self._workspaces is None:
            self._workspaces = WorkspaceList(self)
        return self._workspaces


    def __process_config(self, config: dict):
        """Processes the config dict, whether passed in init or from an API call"""
        self.name = config.get("displayName", "")
        self.address = config.get("address", "")
        self.country = config.get("countryCode", "")
        self.city = config.get("cityName", "")
        self.latitude = config.get("latitude", "")
        self.longitude = config.get("longitude", "")
        self.notes = config.get("notes", "")

    def get_uuid(self):
        """ Returns just the UUID portion of the `id` property.

        .. versionadded:: 4.2.2
            Added because the Base64 ID does not match the :attr:`Device.workspace_location_id`. The assumption is
            that this will be resolved in the future, but needs to be supported until then.

        Returns:
            str: The UUID component of the :attr:`id`

        """
        return decode_spark_id(self.id).split("/")[-1]


class WorkspaceLocationFloor:
    def __init__(self, config: dict):
        """Initialize a new WorkspaceLocationFloor

        Args:
            config (dict): The config as returned by the Webex API

        .. deprecated:: 3.4.0
            The Workspace Location concept is being removed from the Webex APIs. All Workspace Location Floors are now
            :class:`LocationFloor`s now. This class will still show up in come cases until it is completely removed.

        """
        self.name = config.get("displayName")
        self.id = config.get("id")
        self.floor = config.get("floorNumber")
